package org.processmining.logenhancement.abstraction.model;

import org.processmining.logenhancement.abstraction.model.syntax.CompositionCho;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionId;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionInt;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionNCho;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionNPar;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionPar;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionParserVisitor;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepExact;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepInf;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepInfOne;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepZeroOne;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRoot;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionSeq;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.logenhancement.abstraction.model.syntax.Node;
import org.processmining.logenhancement.abstraction.model.syntax.SimpleNode;

public class Printer {

	private static final class CanonicalPrinterVisitor implements CompositionParserVisitor {

		private static StringBuilder sb(Object data) {
			return (StringBuilder) data;
		}

		public Object visit(SimpleNode node, Object data) throws CompositionVisitorException {
			throw new IllegalStateException("No unamed nodes allowed!");
		}

		public Object visit(CompositionRoot node, Object data) throws CompositionVisitorException {
			if (node.jjtGetNumChildren() > 1) {
				throw new IllegalArgumentException("Invalid expression " + node
						+ " should not have been parsed! Top level element is only allowed to have one child.");
			}
			node.jjtGetChild(0).jjtAccept(this, data);
			return null;
		}

		public Object visit(CompositionPar node, Object data) throws CompositionVisitorException {
			printBinaryOp("#", node, data);
			return null;
		}
		
		public Object visit(CompositionNPar node, Object data) throws CompositionVisitorException {
			printNAryOp("#", node, data);
			return null;
		}

		public Object visit(CompositionCho node, Object data) throws CompositionVisitorException {
			printBinaryOp("|", node, data);
			return null;
		}
		
		public Object visit(CompositionNCho node, Object data) throws CompositionVisitorException {
			printNAryOp("|", node, data);
			return null;
		}

		public Object visit(CompositionSeq node, Object data) throws CompositionVisitorException {
			printBinaryOp("<", node, data);
			return null;
		}

		public Object visit(CompositionInt node, Object data) throws CompositionVisitorException {
			printNAryOp("%", node, data);
			return null;
		}
		
		public Object visit(CompositionId node, Object data) throws CompositionVisitorException {
			sb(data).append(node.jjtGetValue());
			return null;
		}

		private void printBinaryOp(String symbol, SimpleNode node, Object data) throws CompositionVisitorException {
			sb(data).append('(');
			node.jjtGetChild(0).jjtAccept(this, data);
			sb(data).append(symbol);
			node.jjtGetChild(1).jjtAccept(this, data);
			sb(data).append(')');
		}
		
		private void printNAryOp(String op, SimpleNode node, Object data) throws CompositionVisitorException {
			sb(data).append(op);
			sb(data).append("[");
			for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
				node.jjtGetChild(i).jjtAccept(this, data);
				if (i != node.jjtGetNumChildren() - 1) {
					sb(data).append(',');
				}
			}
			sb(data).append("]");
		}

		public Object visit(CompositionRepInf node, Object data) throws CompositionVisitorException {
			openParenIfComplex(node.jjtGetChild(0), data);
			node.jjtGetChild(0).jjtAccept(this, data);
			closeParenIfComplex(node.jjtGetChild(0), data);
			sb(data).append("*");
			return null;
		}

		public Object visit(CompositionRepInfOne node, Object data) throws CompositionVisitorException {
			openParenIfComplex(node.jjtGetChild(0), data);
			node.jjtGetChild(0).jjtAccept(this, data);
			closeParenIfComplex(node.jjtGetChild(0), data);
			sb(data).append("+");
			return null;
		}

		public Object visit(CompositionRepZeroOne node, Object data) throws CompositionVisitorException {
			openParenIfComplex(node.jjtGetChild(0), data);
			node.jjtGetChild(0).jjtAccept(this, data);
			closeParenIfComplex(node.jjtGetChild(0), data);
			sb(data).append("?");
			return null;
		}

		public Object visit(CompositionRepExact node, Object data) throws CompositionVisitorException {
			openParenIfComplex(node.jjtGetChild(0), data);
			node.jjtGetChild(0).jjtAccept(this, data);
			closeParenIfComplex(node.jjtGetChild(0), data);
			sb(data).append("{");
			sb(data).append(node.getN());
			sb(data).append(",");
			sb(data).append(node.getM());
			sb(data).append("}");
			return null;
		}

		private void openParenIfComplex(Node node, Object data) {
			if (!isIdentifier(node)) {
				sb(data).append('(');
			}
		}
		private void closeParenIfComplex(Node node, Object data) {
			if (!isIdentifier(node)) {
				sb(data).append(')');
			}
		}

		private boolean isIdentifier(Node node) {
			return node instanceof CompositionId;
		}
		
	}

	private static final CanonicalPrinterVisitor CANONICAL_VISITOR = new CanonicalPrinterVisitor();

	private Printer() {
		super();
	}

	public static String printCanonical(CompositionRoot composition) {
		StringBuilder sb = new StringBuilder();
		try {
			CANONICAL_VISITOR.visit(composition, sb);
		} catch (CompositionVisitorException e) {
			throw new RuntimeException("Exception while trying to print composition formula!", e);
		}
		return sb.toString();
	}

}
