package org.processmining.logenhancement.cpn;

import java.io.InputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.extension.XExtensionManager;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.in.XesXmlParser;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UIImportPlugin;
import org.processmining.framework.abstractplugins.AbstractImportPlugin;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.xeslite.external.XFactoryExternalStore;

/**
 * Taken from KeyValue plug-in and added automatic import of all extra data
 * attributes that happen to be in the cpnxes file.
 * 
 * F. Mannhardt adapted from Michael
 * 
 */
@Plugin(name = "Load CPN XES File (including all data values)", parameterLabels = { "Filename" }, //
		returnLabels = { "Log" }, returnTypes = { XLog.class })
@UIImportPlugin(description = "CPN XES File", extensions = { "cpnxes" }, pack = "LogEnhancement")
public class LoadCPNXESFileWithData extends AbstractImportPlugin {

	private static final String CPNTOOLS_CASE = "cpntools:case";

	private XFactory factory = new XFactoryExternalStore.InMemoryStoreImpl();

	@Override
	protected XLog importFromStream(final PluginContext context, final InputStream input, final String filename,
			final long fileSizeInBytes) throws Exception {
		return importCPNFile(context, input, filename);
	}

	/**
	 * Converts a CPNXES file to a XLog
	 * 
	 * @param context
	 * @param input
	 * @param filename
	 *            may be NULL
	 * @return
	 * @throws Exception
	 */
	public XLog importCPNFile(final PluginContext context, final InputStream input, final String filename)
			throws Exception {
		String modelName = null;
		if (filename != null) {
			modelName = filename.replaceAll("[.]cpnxes$", "").replaceAll(".*[/\\\\]", "");
		}
		context.log("Parsing Log");
		final List<XLog> parse = new XesXmlParser().parse(input);
		if (parse != null && parse.size() >= 1) {
			context.log("Converting to Key/Value Set");
			final KeyValue keyValue = LogToKeyValue.logToKeyValue(context, parse.get(0));

			context.log("Building Mapping");
			final Map<XAttribute, List<String>> mapping = new HashMap<XAttribute, List<String>>();
			XExtension extension;
			int i = 0;
			// Can't find a nicer way to iteratoe over registered extensions :-(
			while ((extension = XExtensionManager.instance().getByIndex(i++)) != null) {
				for (final XAttribute attribute : extension.getEventAttributes()) {
					mapping.put(attribute, Collections.singletonList(attribute.getKey()));
				}
			}

			final Set<String> attributeKeys = new HashSet<String>();
			for (XTrace t : parse.get(0)) {
				for (XEvent e : t) {
					for (XAttribute a : e.getAttributes().values()) {
						if (!attributeKeys.contains(a.getKey()) && !isCPNToolsAttribute(a)) {
							mapping.put(a, Collections.singletonList(a.getKey()));
						}
						attributeKeys.add(a.getKey());
					}
				}
			}

			context.log("Converting Back to Log");
			final Object[] result = KeyValueToLog.keyValueToLog(context, factory,
					new Mapping(Collections.singletonList(CPNTOOLS_CASE), mapping,
							Collections.singletonList(LogToKeyValue.TOTAL_SERIAL)),
					keyValue);

			if (result.length >= 1 && result[0] instanceof XLog) {
				context.getFutureResult(0).setLabel("Log (" + modelName + ")");
				return (XLog) result[0];
			} else {
				context.getFutureResult(0).cancel(true);
				return null;
			}
		}
		context.getFutureResult(0).cancel(true);
		return null;
	}

	private static boolean isCPNToolsAttribute(XAttribute a) {
		return a.getKey().startsWith("cpntools:");
	}
}
