package org.processmining.logenhancement.enriching;

import java.util.Date;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

import com.google.common.collect.Sets;

public class AddDurationBetweenEvents {

	private enum InsertPosition {
		FIRST, SECOND
	}

	@Plugin(name = "Add Elapsed Times between Two Event Classes as Attribute",
			help = "Adds a new attribute (continuous) with the elapsed time between two event classes to either the first of the second event.",
			categories = { PluginCategory.Enhancement }, 
			parameterLabels = { "A log" }, //
	returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void enhanceWithElapsedTime(UIPluginContext context, XLog log) throws UserCancelledException {
		context.getProgress().setMaximum(log.size());

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		XEventClasses eventClasses = XUtils.createEventClasses(classifier, log);
		XFactory factory = XFactoryRegistry.instance().currentDefault();

		Set<XEventClass> eventNames = Sets.newTreeSet(eventClasses.getClasses());

		XEventClass firstEvent = ProMUIHelper.queryForObject(context, "Please choose first event class", eventNames);
		eventNames.remove(firstEvent);
		XEventClass secondEvent = ProMUIHelper.queryForObject(context, "Please choose second event class", eventNames);

		String attributeName = ProMUIHelper.queryForString(context, "Specify attribute name");
		InsertPosition insertPosition = ProMUIHelper.queryForObject(context,
				"Specify to which event the attribute should be added", InsertPosition.values());
		TimeResolution elapsedTimeResolution = ProMUIHelper.queryForObject(context,
				"Specify the resolution to be used for the elapsed time", TimeResolution.values());

		doEnhanceWithElapsedTime(context.getProgress(), log, eventClasses, factory, firstEvent, secondEvent, attributeName,
				insertPosition, elapsedTimeResolution);
	}

	public void doEnhanceWithElapsedTime(Progress progress, XLog log, XEventClasses eventClasses, XFactory factory,
			XEventClass firstEvent, XEventClass secondEvent, String attributeName, InsertPosition insertPosition,
			TimeResolution elapsedTimeResolution) {
		for (XTrace xTrace : log) {
			addElapsedTimeBetweenEventsAsAttr(factory, eventClasses, xTrace, firstEvent, secondEvent,
					attributeName, insertPosition, elapsedTimeResolution);
			progress.inc();
		}
	}

	public static void addElapsedTimeBetweenEventsAsAttr(XFactory factory, XEventClasses eventClasses, XTrace trace,
			XEventClass firstEventClass, XEventClass secondEventClass, String attributeName,
			InsertPosition insertPosition, TimeResolution elapsedTimeResolution) {
		XEvent firstEvent = null;
		XEvent secondEvent = null;
		XAttributeContinuous attrElapsedTime;

		for (XEvent e : trace) {
			if (firstEventClass.equals(eventClasses.getClassOf(e))) {
				firstEvent = e;
			} else {
				if (secondEventClass.equals(eventClasses.getClassOf(e))) {
					if (firstEvent != null) {
						secondEvent = e;
						Date firstDate = XUtils.getTimestamp(firstEvent);
						Date secondDate = XUtils.getTimestamp(secondEvent);
						if (firstDate != null && secondDate != null) {
							long firstTime = firstDate.getTime();
							long secondTime = secondDate.getTime();
							double duration = (secondTime - firstTime) / (double) elapsedTimeResolution.getFactor();
							attrElapsedTime = factory.createAttributeContinuous(attributeName, duration, null);
							if (insertPosition == InsertPosition.FIRST) {
								firstEvent.getAttributes().put(attrElapsedTime.getKey(), attrElapsedTime);
							} else {
								attrElapsedTime = factory.createAttributeContinuous(attributeName, duration, null);
								secondEvent.getAttributes().put(attrElapsedTime.getKey(), attrElapsedTime);
							}
						}
					}
				}
			}
		}
	}

}
