package org.processmining.logenhancement.filtering;

import java.util.Iterator;
import java.util.Set;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.ui.ImprovedEvClassLogMappingUI;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.utils.ProvidedObjectHelper;

import com.google.common.base.Function;
import com.google.common.collect.Collections2;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;

/**
 * Filters out events that cannot be mapped. Copied as the original version does
 * not create a new event log, and does use the
 * {@link ImprovedEvClassLogMappingUI}.
 * 
 * @author F. Mannhardt
 * @author aadrians May 31, 2012 (Original version)
 * 
 */
public class FilterUnmappedEvents {

	@Plugin(name = "Filter unmapped event classes", parameterLabels = { "Petrinet", "Event Logs" }, //
			categories = {
					PluginCategory.Filtering }, help = "Filters out any event that cannot be mapped to one of the activities in the Petri net.", userAccessible = true, returnLabels = {
							"Filtered Log" }, returnTypes = { XLog.class })
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "Felix Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog filter(final UIPluginContext context, Petrinet net, XLog log) throws UserCancelledException {
		ImprovedEvClassLogMappingUI mapping = new ImprovedEvClassLogMappingUI(log, net,
				XUtils.getStandardAndLogDefinedEventClassifiers(log).toArray());
		InteractionResult result = context.showConfiguration("Mapping between Log and Petri net", mapping);

		if (result == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		} else {
			XLog newLog = filter(mapping.getMap(), log);

			context.getFutureResult(0).setLabel(
					ProvidedObjectHelper.getProvidedObjectLabel(context, log) + " - Filtered on Unmapped Events");

			return newLog;
		}
	}

	public XLog filter(TransEvClassMapping mapping, XLog log) {

		Set<String> mappedClassIdentifies = ImmutableSet
				.copyOf(Sets.newHashSet(Collections2.transform(mapping.values(), new Function<XEventClass, String>() {

					public String apply(XEventClass eventClass) {
						return eventClass.getId();
					}
				})));
		XEventClassifier classifier = mapping.getEventClassifier();

		XLog newLog = (XLog) log.clone();

		for (XTrace t : newLog) {
			for (Iterator<XEvent> iterator = t.iterator(); iterator.hasNext();) {
				XEvent event = iterator.next();
				if (!mappedClassIdentifies.contains(classifier.getClassIdentity(event))) {
					iterator.remove();
				}
			}
		}

		return newLog;
	}

}
