package org.processmining.logenhancement.filtering;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.plugins.utils.ProvidedObjectHelper;

public class RemoveEventsFromLogNumberBased {

	@Plugin(name = "Remove random number of events (In Place)", parameterLabels = {}, //
			help = "Removes a random number of events from the input log. Collects all removed events and creates a new log with a single traces that contains the removed events. (The input log is changed!)", categories = {
					PluginCategory.Filtering }, returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void removeRandomEventsFromLog(UIPluginContext context, XLog log) {
		final Progress progress = context.getProgress();
		progress.setMaximum(log.size());

		try {
			int numEvents = ProMUIHelper.queryForInteger(context, "Specify required number of events to be removed");
			XLog[] result = doRemoveRandomEventsFromLog(log, progress, numEvents);

			context.getProvidedObjectManager()
					.createProvidedObject("Filtered " + XConceptExtension.instance().extractName(log) + " @"
							+ DateFormat.getInstance().format(new Date()), result[0], XLog.class, context);
			ProvidedObjectHelper.setFavorite(context, result[0]);

			context.getProvidedObjectManager()
					.createProvidedObject("Removed Events from " + XConceptExtension.instance().extractName(log) + " @"
							+ DateFormat.getInstance().format(new Date()), result[1], XLog.class, context);
			ProvidedObjectHelper.setFavorite(context, result[1]);

		} catch (NumberFormatException e) {
			context.log(e);
		} catch (UserCancelledException e) {
			context.log(e);
		}

	}

	private XLog[] doRemoveRandomEventsFromLog(XLog log, final Progress progress, int numEvents) {
		Random rnd = new Random();
		XLog[] output = new XLog[2];
		Set<Integer> toBeRemoved = new HashSet<Integer>();

		int countEvents = countEvents(log);
		for (int i = 0; i < numEvents; i++) {
			toBeRemoved.add(rnd.nextInt(countEvents));
		}

		// Deep clone original log
		XLog filteredLog = (XLog) log.clone();

		XLog removedEventsLog = XFactoryRegistry.instance().currentDefault().createLog();
		XConceptExtension.instance().assignName(removedEventsLog, "Removed Events");
		XTrace wrapperTrace = XFactoryRegistry.instance().currentDefault().createTrace();
		XConceptExtension.instance().assignName(wrapperTrace, "Wrapper");
		removedEventsLog.add(wrapperTrace);

		int i = 0;
		for (XTrace t : filteredLog) {
			Iterator<XEvent> iter = t.iterator();
			while (iter.hasNext()) {
				XEvent e = iter.next();
				if (toBeRemoved.contains(i)) {
					iter.remove();
					wrapperTrace.add((XEvent) e.clone());
				}
				i++;
			}
			progress.inc();
		}

		output[0] = filteredLog;
		output[1] = removedEventsLog;

		return output;
	}

	private int countEvents(XLog log) {
		int i = 0;
		for (XTrace t : log) {
			i += t.size();
		}
		return i;
	}

	@Plugin(name = "Split event classes from log (In Place)", parameterLabels = {}, //
			returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1, help = "Removes all events of a certain event class from a log. Collects all removed events and creates a new log with a single traces that contains the removed events. (The input log is changed!)")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void removeEventClassesFromLog(UIPluginContext context, XLog log) {
		final Progress progress = context.getProgress();
		progress.setMaximum(log.size());

		try {
			XEventClasses infoEventClasses = XLogInfoFactory.createLogInfo(log).getEventClasses();
			Collection<XEventClass> eventClasses = infoEventClasses.getClasses();
			List<XEventClass> selectedEventClasses = new ArrayList<XEventClass>(eventClasses.size());
			for (XEventClass eventClass : eventClasses) {
				selectedEventClasses.add(eventClass);
			}

			Set<XEventClass> classesToBeRemoved = new HashSet<XEventClass>(ProMUIHelper.queryForObjects(context,
					"Select the event classes to be splitted from log", selectedEventClasses));

			XLog[] result = doRemoveEventClasses(log, progress, infoEventClasses, classesToBeRemoved);

			context.getProvidedObjectManager()
					.createProvidedObject("Filtered " + XConceptExtension.instance().extractName(log) + " @"
							+ DateFormat.getInstance().format(new Date()), result[0], XLog.class, context);
			ProvidedObjectHelper.setFavorite(context, result[0]);

			context.getProvidedObjectManager()
					.createProvidedObject("Removed Events from " + XConceptExtension.instance().extractName(log) + " @"
							+ DateFormat.getInstance().format(new Date()), result[1], XLog.class, context);
			ProvidedObjectHelper.setFavorite(context, result[1]);

		} catch (NumberFormatException e) {
			context.log(e);
		} catch (UserCancelledException e) {
			context.log(e);
		}

	}

	private XLog[] doRemoveEventClasses(XLog log, final Progress progress, XEventClasses infoEventClasses,
			Set<XEventClass> classesToBeRemoved) {

		XLog[] output = new XLog[2];

		// Deep clone original log
		XLog filteredLog = (XLog) log.clone();

		// Create container for new log
		XLog removedEventsLog = XFactoryRegistry.instance().currentDefault().createLog();
		XConceptExtension.instance().assignName(removedEventsLog, "Removed Events");
		XTrace wrapperTrace = XFactoryRegistry.instance().currentDefault().createTrace();
		XConceptExtension.instance().assignName(wrapperTrace, "Wrapper");
		removedEventsLog.add(wrapperTrace);

		for (XTrace t : filteredLog) {
			Iterator<XEvent> iter = t.iterator();
			while (iter.hasNext()) {
				XEvent e = iter.next();
				if (classesToBeRemoved.contains(infoEventClasses.getClassOf(e))) {
					iter.remove();
					wrapperTrace.add((XEvent) e.clone());
				}
			}
			progress.inc();
		}

		output[0] = filteredLog;
		output[1] = removedEventsLog;

		return output;
	}

}
