package org.processmining.logenhancement.filtering.attributes;

import java.util.Iterator;
import java.util.ListIterator;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.Progress;
import org.processmining.logenhancement.filtering.attributes.LogAdapter.NullPlaceHolderAttribute;

/**
 * Manipulate the attributes of events in a log based on attribute name, attribute
 * value, event name.
 * 
 * @author F. Mannhardt
 * 
 */
public class AttributeEditor {

	private LogAdapter logAdapter;
	private AttributeFilterByEnum filterOption;
	private Object filterAttributeValue;
	private String filterAttributeName;
	private XEventClass filterEventClass;

	private Progress progressListener;
	private XAttributeEditor attributeEditor;

	/**
	 * @param progressListener will be used to report progress and incremented for each trace of the log
	 * @param logAdapter
	 * @param filterOption
	 * @param filterAttributeName
	 * @param filterAttributeValue
	 */
	public AttributeEditor(Progress progressListener, LogAdapter logAdapter, AttributeFilterByEnum filterOption,
			String filterAttributeName, Object filterAttributeValue, XAttributeEditor attributeEditor) {
		this.progressListener = progressListener;
		init(logAdapter, filterOption, filterAttributeName, filterAttributeValue, null, attributeEditor);
	}

	/**
	 * @param progressListener  will be used to report progress and incremented for each trace of the log
	 * @param logAdapter
	 * @param filterOption
	 * @param filterAttributeName
	 * @param filterAttributeValue
	 * @param filterEventClass
	 */
	public AttributeEditor(Progress progressListener, LogAdapter logAdapter, AttributeFilterByEnum filterOption,
			String filterAttributeName, Object filterAttributeValue, XEventClass filterEventClass, XAttributeEditor attributeEditor) {
		this.progressListener = progressListener;
		init(logAdapter, filterOption, filterAttributeName, filterAttributeValue, filterEventClass, attributeEditor);
	}

	private void init(LogAdapter logAdapter, AttributeFilterByEnum filterOption, String filterAttributeName,
			Object filterAttributeValue, XEventClass filterEventClass, XAttributeEditor attributeEditor) {
		this.logAdapter = logAdapter;
		this.filterOption = filterOption;
		this.filterAttributeName = filterAttributeName;
		this.filterAttributeValue = filterAttributeValue;
		this.filterEventClass = filterEventClass;
		this.attributeEditor = attributeEditor;
	}

	/**
	 * Filter the Log with the settings that the current instance was created
	 * with.
	 */
	public void processLog() {

		ListIterator<XTrace> traceIterator = logAdapter.getLog().listIterator();
		
		if (filterEventClass == null) {
			if (filterOption == AttributeFilterByEnum.ATTRIBUTE_NAME) {
				filterByName(logAdapter.getLog());
			} else if (filterOption == AttributeFilterByEnum.ATTRIBUTE_NAME_VALUE) {
				filterByNameValue(logAdapter.getLog());
			} else if (filterOption == AttributeFilterByEnum.ATTRIBUTE_VALUE) {
				filterByValue(logAdapter.getLog());
			}
		}
		
		while (traceIterator.hasNext()) {
			XTrace trace = traceIterator.next();
			
			if (filterEventClass == null) {
				if (filterOption == AttributeFilterByEnum.ATTRIBUTE_NAME) {
					filterByName(trace);
				} else if (filterOption == AttributeFilterByEnum.ATTRIBUTE_NAME_VALUE) {
					filterByNameValue(trace);
				} else if (filterOption == AttributeFilterByEnum.ATTRIBUTE_VALUE) {
					filterByValue(trace);
				}
			}

			ListIterator<XEvent> eventIterator = trace.listIterator();
			while (eventIterator.hasNext()) {

				int eventIndex = eventIterator.nextIndex();
				XEvent event = eventIterator.next();

				if (considerEvent(event)) {
					if (filterOption == AttributeFilterByEnum.ATTRIBUTE_NAME) {
						filterByName(event);
					} else if (filterOption == AttributeFilterByEnum.ATTRIBUTE_NAME_VALUE) {
						filterByNameValue(event);
					} else if (filterOption == AttributeFilterByEnum.ATTRIBUTE_VALUE) {
						filterByValue(event);
					}
				}
				trace.set(eventIndex, event);
			}
			progressListener.inc();
		}

	}

	private boolean considerEvent(XEvent oldEvent) {
		if (filterEventClass == null) {
			return true;
		}
		return logAdapter.isEventClass(oldEvent, filterEventClass);
	}

	private void filterByNameValue(XAttributable a) {		
		XAttribute attribute = a.getAttributes().get(filterAttributeName);
		if (attribute != null) {
			processAttribute(a, attribute);
		} else if (filterAttributeValue instanceof NullPlaceHolderAttribute) {
			// Check for add/edit 
			if (!attributeEditor.shouldRemoveAttribute()) {
				a.getAttributes().put(filterAttributeName, attributeEditor.editAttribute(((NullPlaceHolderAttribute) filterAttributeValue).getPrototype()));
			}
			attributeEditor.addAttributes(a);
		}
	}

	private void filterByValue(XAttributable a) {
		if (filterAttributeValue instanceof NullPlaceHolderAttribute) {
			// Check for add/edit 
			if (!attributeEditor.shouldRemoveAttribute()) {
				a.getAttributes().put(filterAttributeName, attributeEditor.editAttribute(((NullPlaceHolderAttribute) filterAttributeValue).getPrototype()));
			}
			attributeEditor.addAttributes(a);
		} else {
			Iterator<XAttribute> iter = a.getAttributes().values().iterator();
			while (iter.hasNext()) {
				processAttribute(a, iter.next());
			}			
		}
	}

	private void processAttribute(XAttributable a, XAttribute attribute) {
		XAttributeMap attributes = a.getAttributes();
		if (attribute instanceof XAttributeLiteral) {
			XAttributeLiteral literal = (XAttributeLiteral) attribute;
			if (considerAttributeByValue(literal)) {
				if (attributeEditor.shouldRemoveAttribute()) {
					attributes.remove(filterAttributeName);	
				} else {
					attributes.put(literal.getKey(), attributeEditor.editAttribute(literal));
				}
				attributeEditor.addAttributes(a);
			}
		} else if (attribute instanceof XAttributeDiscrete) {
			XAttributeDiscrete discrete = (XAttributeDiscrete) attribute;
			if (considerAttributeByValue(discrete)) {
				if (attributeEditor.shouldRemoveAttribute()) {
					attributes.remove(filterAttributeName);	
				} else {
					attributes.put(discrete.getKey(), attributeEditor.editAttribute(discrete));
				}
				attributeEditor.addAttributes(a);
			}
		} else if (attribute instanceof XAttributeContinuous) {
			XAttributeContinuous continuous = (XAttributeContinuous) attribute;
			if (considerAttributeByValue(continuous)) {
				if (attributeEditor.shouldRemoveAttribute()) {
					attributes.remove(filterAttributeName);	
				} else {
					attributes.put(continuous.getKey(), attributeEditor.editAttribute(continuous));
				}
				attributeEditor.addAttributes(a);
			}
		} else if (attribute instanceof XAttributeBoolean) {
			XAttributeBoolean bool = (XAttributeBoolean) attribute;
			if (considerAttributeByValue(bool)) {
				if (attributeEditor.shouldRemoveAttribute()) {
					attributes.remove(filterAttributeName);	
				} else {
					attributes.put(bool.getKey(), attributeEditor.editAttribute(bool));
				}
				attributeEditor.addAttributes(a);
			}
		} else if (attribute instanceof XAttributeTimestamp) {
			XAttributeTimestamp time = (XAttributeTimestamp) attribute;
			if (considerAttributeByValue(time)) {
				if (attributeEditor.shouldRemoveAttribute()) {
					attributes.remove(filterAttributeName);	
				} else {
					attributes.put(time.getKey(), attributeEditor.editAttribute(time));
				}
				attributeEditor.addAttributes(a);
			}
		} else {
			throw new UnsupportedOperationException("No support for datatype of " + attribute);
		}
	}

	private boolean considerAttributeByValue(Object value) {
		return filterAttributeValue.equals(value);
	}

	private void filterByName(XAttributable a) {
		XAttributeMap attributes = a.getAttributes();
		if (attributeEditor.shouldRemoveAttribute()) {
			attributes.remove(filterAttributeName);	
		} else {
			XAttribute attribute = attributes.get(filterAttributeName);
			if (attribute != null) {
				attributes.put(attribute.getKey(), attributeEditor.editAttribute(attribute));
			}
		}
		attributeEditor.addAttributes(a);
	}

}
