package org.processmining.logenhancement.filtering.attributes;

import java.util.Comparator;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;

/**
 * Provides information about the Log
 * 
 * @author fmannhardt
 * 
 */
public class LogAdapter {

	public interface NullPlaceHolderAttribute {

		XAttribute getPrototype();

	}

	public abstract static class NullPlaceHolderAttributeImpl implements NullPlaceHolderAttribute {

		public String toString() {
			return "NULL (Attribute missing)";
		}

	};

	private final XLog log;
	private final XEventClassifier classifier;

	private XLogInfo logInfo;

	public LogAdapter(XLog log, XEventClassifier classifier) {
		super();
		this.log = log;
		this.classifier = classifier;
	}

	public Set<String> getAllEventAttributes() {
		Set<String> attributeNames = new TreeSet<String>();
		for (XTrace trace : log) {
			for (XEvent event : trace) {
				XAttributeMap attributes = event.getAttributes();
				attributeNames.addAll(attributes.keySet());
			}
		}
		return attributeNames;
	}

	public Set<Object> getAllEventValues(String selectedAttribute) {
		TreeSet<Object> values = new TreeSet<Object>(new Comparator<Object>() {

			public int compare(Object o1, Object o2) {
				if (o1 instanceof NullPlaceHolderAttribute && o2 instanceof NullPlaceHolderAttribute) {
					return 0;
				}
				if (o1 instanceof NullPlaceHolderAttribute) {
					return -1;
				}
				if (o2 instanceof NullPlaceHolderAttribute) {
					return 1;
				}
				return ((Comparable) o1).compareTo(o2);
			}
		});
		XAttribute currentAttribute = null;
		boolean hasMissingValues = false;
		for (XTrace trace : log) {
			for (XEvent event : trace) {
				XAttributeMap attributes = event.getAttributes();
				if (attributes.containsKey(selectedAttribute)) {
					currentAttribute = attributes.get(selectedAttribute);
					values.add(currentAttribute);
				} else {
					hasMissingValues = true;
				}
			}
		}
		if (hasMissingValues) {
			assert currentAttribute != null : "We should have seen at least one attribute with key" + selectedAttribute;
			NullPlaceHolderAttribute noValue = getNoValuePlaceholder(currentAttribute);
			values.add(noValue);
		}
		return values;
	}

	private static NullPlaceHolderAttribute getNoValuePlaceholder(final XAttribute a) {
		return new NullPlaceHolderAttributeImpl() {

			public XAttribute getPrototype() {
				return (XAttribute) a.clone();
			}

		};
	}

	public Set<XEventClass> getAllEventClasses() {
		return new TreeSet<XEventClass>(getLogInfo().getEventClasses().getClasses());
	}

	private XLogInfo getLogInfo() {
		if (logInfo == null) {
			if (classifier == null) {
				logInfo = XLogInfoFactory.createLogInfo(log);
			} else {
				logInfo = XLogInfoFactory.createLogInfo(log, classifier);
			}
		}
		return logInfo;
	}

	public boolean isEventClass(XEvent e, XEventClass eventClass) {
		return eventClass.equals(getLogInfo().getEventClasses().getClassOf(e));
	}

	public int size() {
		return log.size();
	}

	public XLog getLog() {
		return log;
	}

}
