package org.processmining.logenhancement.merge.alignmentbased;

import java.text.MessageFormat;
import java.util.NavigableSet;
import java.util.Set;

import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizerException;
import org.processmining.plugins.balancedconformance.result.BalancedReplayResult;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;

/**
 * Plug-in that merges a set of events with a "main" event log based on the
 * alignment of a DPN-net and the "main" event log. As described in the BPI'14
 * paper: <a href="http://dx.doi.org/10.1007/978-3-319-15895-2_21">Extending
 * Process Logs with Events from Supplementary Sources</a>
 * 
 * @author F. Mannhardt
 * 
 */
public class AlignmentBasedLogEnhancementPlugin {

	private final class DefaultMergeObserver implements MergeObserver {

		private final PluginContext context;
		private final boolean isDebug;
		private final Progress progress;
		private int progressWindow;
		private int mergedTraces;
		private int alignmentCount;

		public DefaultMergeObserver(PluginContext context, boolean isDebug) {
			this.context = context;
			this.isDebug = isDebug;
			this.progress = context.getProgress();
		}

		public void processingLog(XLog masterLog, XLogInfo masterLogInfo, Set<XEvent> availableEvents,
				MergeConfiguration config) {
			this.progress.setMinimum(0);
			this.progress.setValue(0);
			this.progressWindow = Math.max(masterLog.size() / 25, 10);
			this.progress.setMaximum(((masterLog.size() / progressWindow) * 2) + 2);
		}

		public void startingAlignmentPhase(PetriNetWithData net, XLog masterLog) {

		}

		public void receivedAlignment(XTrace masterTrace, XTrace enrichedTrace, DataAlignedTrace dataAlignment) {
			if ((++alignmentCount) % progressWindow == 0) {
				progress.inc();
			}
		}

		public void finishedAlignmentPhase(PetriNetWithData net, XLog masterLog, BalancedReplayResult replayResult) {

		}

		public void startMergeRound(int round, int currentMatchLimit, NavigableSet<Alignment> resultSet,
				int retrySetSize, Set<XEvent> eventSet) {
			log("Start merge round {0} with match limit {1}", round, currentMatchLimit);
			log("Still {0} traces to enhance and {1} events to merge", retrySetSize, eventSet.size());
		}

		public void mergedTrace(LogEnhancementResult mergeResult) {
			mergedTraces++;
			if (mergedTraces % progressWindow == 0) {
				progress.inc();
			}
		}

		public void finishedMergeLog(XLog masterLog, XLog mergedLog) {
			log("Finished enhancing log");
		}

		public void log(String message, Throwable e) {
			context.log(message, MessageLevel.ERROR);
			context.log(e);
		}

		public void log(String message, Object... args) {
			context.log(MessageFormat.format(message, args));
		}

		public void debug(String msg, Object... args) {
			if (isDebug) {
				context.log(MessageFormat.format(msg, args));
			}
		}

	}

	@Plugin(name = "Extend/Enhance Event Log with Supplementary Events", parameterLabels = { "DPN-net",
			"Event Log (Master)", "Event Logs (Additional)" }, returnLabels = { "Merged Event Log" }, //
	returnTypes = { XLog.class }, userAccessible = true, mostSignificantResult = 1, //
	help = "Extends a master event log with supplementary events from multiple additional event logs based on the alignment between a DPN-net and the event logs."
			+ " The technique is described in the BPI'14 paper Extending Process Logs with Events from Supplementary Sources (DOI: 10.1007/978-3-319-15895-2_21)", categories = { PluginCategory.Enhancement })
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog mergeLogWithProgress(UIPluginContext context, PetriNetWithData net, XLog masterLog,
			XLog... additionalLogs) throws DataAlignmentException, StringDiscretizerException {
		try {
			MergeConfiguration configuration = AlignmentBasedLogEnhancer.queryConfiguration(context, net, masterLog,
					additionalLogs);
			return doMerge(context, configuration, net, masterLog, additionalLogs);
		} catch (UserCancelledException e) {
			context.getFutureResult(0).cancel(true);
			return null;
		} catch (LogEnhancementException e) {
			context.log(e);
			context.getFutureResult(0).cancel(true);
			return null;
		}
	}

	public XLog doMerge(PluginContext context, MergeConfiguration configuration, PetriNetWithData net, XLog masterLog,
			XLog[] additionalLogs) throws LogEnhancementException, DataAlignmentException, StringDiscretizerException {
		return doMerge(context.getProgress(), configuration, new DefaultMergeObserver(context, false), net, masterLog,
				additionalLogs);
	}

	public XLog doMerge(Progress progress, MergeConfiguration configuration, MergeObserver observer,
			PetriNetWithData net, XLog masterLog, XLog[] additionalLogs) throws LogEnhancementException,
			DataAlignmentException, StringDiscretizerException {
		return new AlignmentBasedLogEnhancer(configuration, net, masterLog, additionalLogs, observer, progress)
				.doMergeLog();
	}

}
