package org.processmining.logenhancement.merge.alignmentbased;

import java.text.NumberFormat;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.NavigableSet;
import java.util.TreeSet;

import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryNaiveImpl;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.ui.ImprovedEvClassLogMappingUI;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.logenhancement.merge.alignmentbased.TraceSet.Candidates;
import org.processmining.logenhancement.merge.alignmentbased.TraceSet.TraceEntry;
import org.processmining.logenhancement.merge.alignmentbased.dataconstraints.DataAlignmentConstraintsChecker;
import org.processmining.logenhancement.merge.alignmentbased.timeconstraints.TimeConstraints;
import org.processmining.logenhancement.merge.alignmentbased.timeconstraints.TimeConstraintsChecker;
import org.processmining.logenhancement.merge.alignmentbased.timeconstraints.TimeConstraintsCheckerStructureImpl;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.GUI.MappingPanel;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.plugins.balancedconformance.AbstractBalancedDataConformancePlugin;
import org.processmining.plugins.balancedconformance.BalancedDataConformancePlusPlugin;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactoryImpl;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizerException;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizerImpl;
import org.processmining.plugins.balancedconformance.mapping.Variable;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserverNoOpImpl;
import org.processmining.plugins.balancedconformance.result.BalancedReplayResult;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.base.Function;
import com.google.common.collect.Maps;

import gnu.trove.map.TIntObjectMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.procedure.TIntObjectProcedure;

/**
 * Merges a set of events with a "main" event log based on the alignment of a
 * DPN-net and the "main" event log.
 * 
 * @author F. Mannhardt
 * 
 */
public class AlignmentBasedLogEnhancer {

	private static final class MergeAlignmentObserver extends DataConformancePlusObserverNoOpImpl {

		private final MergeObserver observer;
		private DescriptiveStatistics timeStats;
		private DescriptiveStatistics queueStats;

		public MergeAlignmentObserver(MergeObserver observer) {
			this.observer = observer;
			this.timeStats = new DescriptiveStatistics(100);
			this.queueStats = new DescriptiveStatistics(100);
		}

		public void foundOptimalAlignment(int traceIndex, XTrace trace, DataAlignedTrace dataAlignment,
				int partialDataAlignmentsNeeded, int cacheHit, int cacheSize, long queuedStates, long dataStateCount,
				long usedTime) {
			timeStats.addValue(usedTime);
			queueStats.addValue(queuedStates);
			if ((dataStateCount % 100000) == 0) {
				log("Data States in Memory: " + dataStateCount);
				NumberFormat nf = NumberFormat.getNumberInstance();
				log("Average A* Queue Size per Trace: " + nf.format(queueStats.getMean()));
				log("Average Time per Trace: " + nf.format(timeStats.getMean()));
			}
			observer.receivedAlignment(trace, trace, dataAlignment);
		}

		public void log(String message) {
			observer.log(message);
		}

		public void log(String message, Throwable e) {
			observer.log(message, e);
		}

	}

	private final MergeConfiguration config;
	private final PetriNetWithData net;
	private final XLog masterLog;
	private final XFactory xFactory;
	private final MergeAlignmentObserver alignmentObserver;
	private final MergeObserver mergeObserver;
	private final Progress progressListener;
	private final Map<String, GuardExpression> guardMap;

	private final LogMapping logMapping;
	private final StringDiscretizer discretizer;

	private final XLogInfo masterLogInfo;

	private final EventSet eventSet;
	private final TraceSet traceSet;

	private DataStateFactory stateFactoy;

	private int checkCount;

	private final ComparisonEventEntry earliestEvent;
	private final ComparisonEventEntry latestEvent;
	private final ComparisonEventEntry idealEvent;

	@SuppressWarnings("rawtypes")
	public AlignmentBasedLogEnhancer(MergeConfiguration configuration, PetriNetWithData net, XLog masterLog,
			XLog[] additionalLogs, MergeObserver observer, Progress progressListener)
			throws DataAlignmentException, StringDiscretizerException {
		this.config = configuration;
		this.net = net;
		this.masterLog = masterLog;
		this.mergeObserver = observer;
		this.progressListener = progressListener;
		this.xFactory = XFactoryRegistry.instance().currentDefault();
		this.alignmentObserver = new MergeAlignmentObserver(observer);
		this.config.conformanceConfig.setObserver(alignmentObserver);
		this.guardMap = createGuardMap(net);

		logMapping = new LogMapping(config.conformanceConfig.getActivityMapping(),
				config.conformanceConfig.getMapEvClass2Cost(), config.conformanceConfig.getMapTrans2Cost(),
				config.conformanceConfig.getVariableMapping(), config.conformanceConfig.getVariableCost(),
				config.conformanceConfig.getUpperBounds(), config.conformanceConfig.getLowerBounds(),
				config.conformanceConfig.getDefaultValues(), config.conformanceConfig.getVariablesUnassignedMode(), net,
				joinLogs(additionalLogs));
		this.discretizer = new StringDiscretizerImpl(joinLogs(masterLog, additionalLogs), guardMap.values(),
				config.extraVariablesMapping, LogMapping.convertToVariableTypes(logMapping.getVariables()),
				config.conformanceConfig.getDefaultValues(), logMapping.getEventIdentityToTransition(),
				logMapping.getVariablesToWrite(), config.conformanceConfig.getActivityMapping().getEventClassifier());

		// Use user selected EventClassifier for LogInfo		
		XEventClassifier selectedClassifier = config.conformanceConfig.getActivityMapping().getEventClassifier();
		this.masterLogInfo = XLogInfoFactory.createLogInfo(masterLog, selectedClassifier);

		mergeObserver.log("Creating Initial Datastructures ...");

		this.stateFactoy = new DataStateFactoryImpl(
				Maps.transformValues(logMapping.getVariables(), new Function<Variable, Class>() {

					public Class apply(Variable variable) {
						return variable.getType();
					}
				}));
		this.eventSet = new EventSetImpl(stateFactoy, LogMapping.convertToVariableTypes(logMapping.getVariables()),
				config.extraVariablesMapping, discretizer, config.extraEventsMapping.getEventClassifier(),
				additionalLogs);
		this.traceSet = new TraceSetImpl(masterLog);
		this.earliestEvent = new ComparisonEventEntry(Integer.MIN_VALUE);
		this.latestEvent = new ComparisonEventEntry(Integer.MAX_VALUE);
		this.idealEvent = new ComparisonEventEntry(Integer.MIN_VALUE);

		mergeObserver.log("Finished!");
	}

	private static Map<String, GuardExpression> createGuardMap(PetriNetWithData net) {
		Map<String, GuardExpression> guardMap = new HashMap<>();
		for (Transition t : net.getTransitions()) {
			if (((PNWDTransition) t).getGuardExpression() != null) {
				guardMap.put(t.getLabel(), ((PNWDTransition) t).getGuardExpression());
			}
		}
		return guardMap;
	}

	public static MergeConfiguration queryConfiguration(UIPluginContext context, PetriNetWithData net, XLog masterLog,
			XLog... additionalLogs) throws UserCancelledException {
		MergeConfiguration configuration = new MergeConfiguration();

		context.showConfiguration("Configure Alignment Phase (Using Balanced Conformance Checking of DPN)",
				SlickerFactory.instance().createLabel("--->"));

		configuration.conformanceConfig = AbstractBalancedDataConformancePlugin.queryConfiguration(context, net,
				masterLog);
		// Just a shortcut
		configuration.eventClassifier = configuration.conformanceConfig.getActivityMapping().getEventClassifier();

		context.showConfiguration("Configure Enhancement/Extension Phase",
				SlickerFactory.instance().createLabel("--->"));

		XLog combinedExtraLogs = joinLogs(additionalLogs);
		ImprovedEvClassLogMappingUI mappingUi = new ImprovedEvClassLogMappingUI(combinedExtraLogs, net,
				new Object[] { configuration.eventClassifier });
		InteractionResult result = context.showConfiguration("Mapping of Additional Events to Transitions in the DPN",
				mappingUi);
		if (result == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		} else {
			configuration.extraEventsMapping = mappingUi.getMap();
		}

		XLog combinedAllLogs = joinLogs(masterLog, additionalLogs);
		Collection<String> logAttributes = XLogInfoFactory.createLogInfo(combinedAllLogs, configuration.eventClassifier)
				.getEventAttributeInfo().getAttributeKeys();
		TreeSet<String> processAttributes = new TreeSet<>();
		for (DataElement variable : net.getVariables()) {
			processAttributes.add(variable.getVarName());
		}

		MappingPanel<String, String> mapVariablePanel = new MappingPanel<>(processAttributes, logAttributes);
		result = context.showConfiguration("Mapping of Attributes in the Additional Events to Variables in the DPN",
				mapVariablePanel);
		if (result == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		} else {
			configuration.extraVariablesMapping = mapVariablePanel.getMapping(false);
		}

		try {
			int maxMatches = ProMUIHelper.queryForInteger(context, "Matching Limit");
			configuration.maxMatchesLimit = maxMatches;
		} catch (NumberFormatException e) {
			context.log(e);
			throw new UserCancelledException(e);
		} catch (UserCancelledException e) {
			throw new UserCancelledException(e);
		}

		try {
			int timeLimit = ProMUIHelper.queryForInteger(context, "Time Horizon Limit (in days)");
			configuration.startDateLimit = timeLimit * 1000l * 60l * 60l * 24l;
			configuration.endDateLimit = timeLimit * 1000l * 60l * 60l * 24l;
		} catch (NumberFormatException e) {
			context.log(e);
			throw new UserCancelledException(e);
		} catch (UserCancelledException e) {
			throw new UserCancelledException(e);
		}

		return configuration;
	}

	public XLog doMergeLog() throws LogEnhancementException, DataAlignmentException {

		mergeObserver.processingLog(masterLog, masterLogInfo, eventSet.asSet(), config);

		XLog mergedLog = xFactory.createLog((XAttributeMap) masterLog.getAttributes().clone());

		try {

			mergeObserver.startingAlignmentPhase(net, masterLog);

			BalancedDataConformancePlusPlugin balancedConformancePlugin = new BalancedDataConformancePlusPlugin();
			BalancedReplayResult replayResult = balancedConformancePlugin.doBalancedAlignmentDataConformanceChecking(
					net, masterLog, progressListener, config.conformanceConfig);

			mergeObserver.finishedAlignmentPhase(net, masterLog, replayResult);

			// 1. Round: Start with highest fitness
			NavigableSet<Alignment> resultSet = replayResult.labelStepArray.descendingSet();

			mergeObserver.startMergeRound(1, config.maxMatchesFirstRound, resultSet, traceSet.retryCount(),
					eventSet.asSet());

			for (Alignment alignment : resultSet) {

				XTrace masterTrace = traceSet.getByName(alignment.getTraceName());
				Candidates candidates = mergeTraceByAlignment(masterTrace, masterLogInfo, alignment,
						config.maxMatchesFirstRound);

				if (candidates.isShouldRetry()) {
					traceSet.markRetry(masterTrace, alignment, candidates);
				} else {
					LogEnhancementResult mergeResult = doMergeTrace(masterTrace, alignment,
							candidates.createResult(eventSet));
					mergeObserver.mergedTrace(mergeResult);
					mergedLog.add(mergeResult.getMergedTrace());
				}
			}

			int currentLimit = config.maxMatchesFirstRound;
			int round = 2;

			// 2.-n. Round: Retry ambiguous matches
			while (traceSet.hasRetries() && currentLimit < config.maxMatchesLimit) {

				mergeObserver.startMergeRound(round, currentLimit, resultSet, traceSet.retryCount(), eventSet.asSet());

				// Increase limit
				currentLimit = 2 * currentLimit;

				mergeObserver.debug(
						"Retrying {0} ambiguous matches, just processing matches with less than {1} candidates",
						traceSet.retryCount(), currentLimit);

				for (Iterator<TraceEntry> iterator = traceSet.retryIterator(); iterator.hasNext();) {

					TraceEntry retryEntry = iterator.next();
					XTrace masterTrace = retryEntry.getTrace();
					Alignment alignment = retryEntry.getAlignment();

					Candidates candidates = decideCandidates(masterTrace, alignment, currentLimit,
							retryEntry.getCandidates());

					if (candidates.isShouldRetry()) {
						mergeObserver.debug("Keeping {0} in retry set", retryEntry);
						retryEntry.setCandidates(candidates);
					} else {
						// Remove from future retries
						iterator.remove();
						LogEnhancementResult mergeResult = doMergeTrace(masterTrace, alignment,
								candidates.createResult(eventSet));
						mergeObserver.mergedTrace(mergeResult);
						mergedLog.add(mergeResult.getMergedTrace());
					}

				}
				round++;
			}

		} catch (ControlFlowAlignmentException e) {
			throw new LogEnhancementException("Alignment Error", e);
		}

		// Report the rest of the non-merged traces
		for (Iterator<TraceEntry> iterator = traceSet.retryIterator(); iterator.hasNext();) {
			TraceEntry retryEntry = iterator.next();
			// Create empty result
			LogEnhancementResult mergeResult = doMergeTrace(retryEntry.getTrace(), retryEntry.getAlignment(),
					new TIntObjectHashMap<EventEntry>());
			mergeObserver.mergedTrace(mergeResult);
			mergedLog.add(mergeResult.getMergedTrace());
		}

		mergeObserver.finishedMergeLog(masterLog, mergedLog);

		return mergedLog;
	}

	private Candidates mergeTraceByAlignment(XTrace masterTrace, XLogInfo masterLogInfo, Alignment alignment,
			int maxMatches) {

		// Use the discretizer which is aware of the extra events
		DataAlignmentConstraintsChecker dataChecker = new DataAlignmentConstraintsChecker(stateFactoy, discretizer,
				config.extraVariablesMapping, LogMapping.convertToVariableTypes(logMapping.getVariables()), alignment);

		TimeConstraintsChecker timeChecker = new TimeConstraintsCheckerStructureImpl(net,
				config.conformanceConfig.getInitialMarking(), config.conformanceConfig.getActivityMapping(),
				masterTrace, alignment, config.startDateLimit, config.endDateLimit);

		// Phase 1: Find possible event candidates
		Candidates mergeCandidates = findCandidates(masterTrace, alignment, dataChecker, timeChecker);

		// Phase 2: Decide which candidate to merge in case of multiple candidates
		return decideCandidates(masterTrace, alignment, maxMatches, mergeCandidates);
	}

	private LogEnhancementResult doMergeTrace(XTrace masterTrace, Alignment alignment,
			TIntObjectMap<EventEntry> eventsToMerge) {
		XTrace mergedTrace = cloneWithFactory(xFactory, masterTrace);
		for (XEvent e : masterTrace) {
			// Already in correct order
			mergedTrace.add(e);
		}

		LogEnhancementResultImpl alignmentBasedMergeResult = new LogEnhancementResultImpl();

		for (EventEntry e : eventsToMerge.valueCollection()) {
			// Add in correct order
			mergedTrace.insertOrdered(e.getEvent());
			// Remove from available 
			eventSet.markUnavailable(e);
		}

		alignmentBasedMergeResult.setAlignment(alignment);
		alignmentBasedMergeResult.setMasterTrace(masterTrace);
		alignmentBasedMergeResult.setMergedEvents(eventsToMerge);
		alignmentBasedMergeResult.setMergedTrace(mergedTrace);

		return alignmentBasedMergeResult;
	}

	private Candidates findCandidates(XTrace masterTrace, Alignment alignment,
			DataAlignmentConstraintsChecker dataChecker, TimeConstraintsChecker timeChecker) {

		Candidates candidates = new Candidates();
		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator();

		while (stepIter.hasNext()) {
			int currentIndex = stepIter.nextIndex();
			DataAwareStepTypes stepType = stepIter.next();
			if (stepType == DataAwareStepTypes.MREAL) {
				ExecutionStep processStep = alignment.getProcessTrace().get(currentIndex);
				addMatchingEvents(candidates, masterTrace, currentIndex, processStep, dataChecker, timeChecker);
			}
		}

		return candidates;
	}

	private Candidates decideCandidates(final XTrace masterTrace, final Alignment alignment, final int maxMatches,
			final Candidates oldCandidates) {

		final Candidates newCandidates = new Candidates();
		final TIntObjectMap<BitSet> mergeCandidates = oldCandidates.getEventMatches();
		final TIntObjectMap<TimeConstraints> timeConstraints = oldCandidates.getTimeConstraints();

		mergeCandidates.forEachEntry(new TIntObjectProcedure<BitSet>() {

			public final boolean execute(int index, BitSet eventCandidates) {

				removeMergedEvents(eventCandidates);

				int candidateCount = eventCandidates.cardinality();

				if (candidateCount == 1) {

					mergeObserver.debug("Found perfect match for {0}",
							XConceptExtension.instance().extractName(masterTrace));
					newCandidates.addMatch(index, eventCandidates.nextSetBit(0));

				} else if (candidateCount > 1 && candidateCount <= maxMatches) {

					mergeObserver.debug("Choosing from {0} matches for {1}", candidateCount,
							XConceptExtension.instance().extractName(masterTrace));

					List<EventEntry> sortedEvents = eventSet.getEventsSortedByTime(eventCandidates);

					TimeConstraints currentConstraint = timeConstraints.get(index);

					//TODO replace by flexible strategy pattern
					long timeRange = currentConstraint.getTimeAfter().getTime()
							- currentConstraint.getTimeBefore().getTime();
					long timeInBetween = currentConstraint.getTimeBefore().getTime() + (timeRange / 2);
					Date wantedTime = new Date(timeInBetween);

					idealEvent.setDate(wantedTime);

					// This assumes that the event list is sorted by time
					int searchResult = Collections.binarySearch(sortedEvents, idealEvent);

					if (searchResult < 0) {
						// look for nearest match
						searchResult = ~searchResult;

						if (searchResult == sortedEvents.size()) {
							// Everything is smaller, return last event
							newCandidates.addMatch(index, sortedEvents.get(searchResult - 1).getEventIndex());
						} else if (searchResult == 0) {
							// Everything is bigger, return first event
							newCandidates.addMatch(index, sortedEvents.get(0).getEventIndex());
						} else {
							// Look which entry is closer
							EventEntry larger = sortedEvents.get(searchResult);
							EventEntry smaller = sortedEvents.get(searchResult - 1);

							long largerDiff = Math.abs(larger.getDate().getTime() - wantedTime.getTime());
							long smallerDiff = Math.abs(smaller.getDate().getTime() - wantedTime.getTime());

							if (largerDiff < smallerDiff) {
								newCandidates.addMatch(index, larger.getEventIndex());
							} else {
								newCandidates.addMatch(index, smaller.getEventIndex());
							}
						}
					} else {
						// perfect match
						newCandidates.addMatch(index, sortedEvents.get(searchResult).getEventIndex());
					}

				} else if (candidateCount > maxMatches) {
					mergeObserver.debug("Too many matches ({0}) for {1}", eventCandidates.size(),
							XConceptExtension.instance().extractName(masterTrace));
					newCandidates.setEventMatches(oldCandidates.getEventMatches());
					newCandidates.setTimeConstraints(oldCandidates.getTimeConstraints());
					newCandidates.setShouldRetry(true);
				} else {
					mergeObserver.debug("Could not find any matches for {0}",
							XConceptExtension.instance().extractName(masterTrace));
				}

				return true;
			}

			private final void removeMergedEvents(BitSet eventCandidates) {
				eventCandidates.andNot(eventSet.getUsedEvents());
			}

		});

		return newCandidates;
	}

	private final void addMatchingEvents(Candidates candidates, XTrace masterTrace, int currentIndex,
			ExecutionStep modelMove, DataAlignmentConstraintsChecker dataChecker, TimeConstraintsChecker timeChecker) {

		assert modelMove.getActivityObject() != null
				&& modelMove.getActivityObject() instanceof Transition : "Missing Transition for ExecutionStep "
						+ modelMove;

		final Transition neededTransition = (Transition) modelMove.getActivityObject();
		XEventClass neededEventClass = config.extraEventsMapping.get(neededTransition);
		if (neededEventClass == config.extraEventsMapping.getDummyEventClass()) {
			// This event is not part of the additional events, stop trying
			return;
		}

		TimeConstraints timeConstraints = timeChecker.getTimeConstraints(currentIndex, modelMove);
		candidates.addTimeConstraint(currentIndex, timeConstraints);
		earliestEvent.setDate(timeConstraints.getTimeBefore());
		latestEvent.setDate(timeConstraints.getTimeAfter());

		assert earliestEvent.getDate().before(latestEvent.getDate())
				|| earliestEvent.getDate().equals(latestEvent.getDate()) : "Invalid time constraint!";

		NavigableSet<EventEntry> events = eventSet.findByClass(neededEventClass).subSet(earliestEvent, true,
				latestEvent, true);

		for (EventEntry e : events) { // Iterate in ascending order of the time stamp
			// Check if the data in the event matches the alignment
			if (dataChecker.checkConstraints(currentIndex, e)) {
				// Add event index
				candidates.addMatch(currentIndex, e.getEventIndex());
			}

			checkCount++;
			if (checkCount % 10000000 == 0) {
				mergeObserver.log("Already checked {0} combinations of event candidates and model-moves.", checkCount);
			}
		}
	}

	private static XLog joinLogs(XLog log, XLog... logs) {
		XLog[] combined = new XLog[1 + logs.length];
		combined[0] = log;
		for (int i = 0; i < logs.length; i++) {
			combined[i + 1] = logs[i];
		}
		return joinLogs(combined);
	}

	private static XLog joinLogs(XLog[] logs) {
		XFactory factory = new XFactoryNaiveImpl();
		XLog combinedLog = factory.createLog();
		for (XLog log : logs) {
			combinedLog.addAll(log);
		}
		return combinedLog;
	}

	private XTrace cloneWithFactory(XFactory xFactory, XTrace t) {
		return xFactory.createTrace((XAttributeMap) t.getAttributes().clone());
	}

}
