package org.processmining.logenhancement.merge.alignmentbased.dataconstraints;

import gnu.trove.map.TIntObjectMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.THashSet;

import java.text.MessageFormat;
import java.util.ArrayDeque;
import java.util.Date;
import java.util.Deque;
import java.util.HashSet;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.syntax.ParseException;
import org.processmining.logenhancement.merge.alignmentbased.EventEntry;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.guards.Expression;
import org.processmining.models.guards.NumericValueConversion;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.util.ExpressionEvaluator;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;

public class DataAlignmentConstraintsChecker {

	public static class DataCheck {

		/**
		 * Mapping from variable name in DPN-net to its type
		 */
		@SuppressWarnings("rawtypes")
		private final Map<String, Class> varType;

		private final StringDiscretizer discretizer;
		private final Alignment alignment;
		private final GuardExpression combinedExpression;
		private final THashSet<DataState> fulfilledStates = new THashSet<>(); 		

		public DataCheck(Alignment alignment, StringDiscretizer discretizer,
				@SuppressWarnings("rawtypes") Map<String, Class> varType, GuardExpression combinedExpression) {
			this.alignment = alignment;
			this.discretizer = discretizer;
			this.varType = varType;
			this.combinedExpression = combinedExpression;
		}

		public boolean check(int index, EventEntry eventCandidate) {
			ExecutionStep processStep = alignment.getProcessTrace().get(index);
			if (eventCandidate.getAttributes().isEmpty() && processStep.isEmpty()) {
				// No write operations -> nothing to check
				return true;
			} else {
				if (combinedExpression != null) {
					DataState eventState = eventCandidate.getAttributes();
					if (fulfilledStates.contains(eventState)) {
						return true;
					}
					boolean guardsFulfilled = ExpressionEvaluator.isTrue(combinedExpression, eventState, varType,
							discretizer);
					if (guardsFulfilled) {
						fulfilledStates.add(eventCandidate.getAttributes());
					}
					return guardsFulfilled;
				} else {
					return true;
				}
			}
		}

	}

	private static final boolean IS_DEBUG = false;

	private final TIntObjectMap<DataCheck> dataCheckMap;
	private final Map<String, Class> varType;
	private final StringDiscretizer discretizer;

	public DataAlignmentConstraintsChecker(DataStateFactory stateFactoy, StringDiscretizer discretizer,
			Map<String, String> variableMapping, @SuppressWarnings("rawtypes") Map<String, Class> varType,
			Alignment alignment) {

		this.discretizer = discretizer;
		this.varType = varType;
		this.dataCheckMap = new TIntObjectHashMap<>();

		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator();
		ListIterator<ExecutionStep> processIter = alignment.getProcessTrace().listIterator();
		DataState currentDataState = stateFactoy.newDataState();

		while (stepIter.hasNext()) {
			assert processIter.hasNext();
			DataAwareStepTypes stepType = stepIter.next();
			ExecutionStep processStep = processIter.next();
			int currentIndex = stepIter.previousIndex();

			if (stepType != DataAwareStepTypes.L && stepType != DataAwareStepTypes.MINVI) {

				if (stepType == DataAwareStepTypes.MREAL) {
					Expression oldExpr = generateExpression(currentIndex, currentDataState, alignment,
							processStep);
					try {
						GuardExpression combinedExpression =  GuardExpression.Factory.newInstance(oldExpr.toString()) ;
						dataCheckMap.put(currentIndex, new DataCheck(alignment, discretizer, varType, combinedExpression));
					} catch (ParseException e) {
						throw new RuntimeException(e);
					}
				}

				for (Entry<String, Object> entry : processStep.entrySet()) {
					Object value = entry.getValue();
					String varName = entry.getKey();
					DataStateHelper.storeValue(currentDataState, varName, value, value.getClass(), discretizer);
				}

			}
		}
	}

	private Expression generateExpression(int currentIndex, DataState currentDataState, Alignment alignment,
			ExecutionStep modelMove) {

		// Start at the model move index
		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator(currentIndex);
		ListIterator<ExecutionStep> processIter = alignment.getProcessTrace().listIterator(currentIndex);

		Set<String> writeOps = new HashSet<>();
		for (String varName : modelMove.keySet()) {
			writeOps.add(varName.concat("'"));
		}

		debug("Current Variables {0}", writeOps);

		Deque<Expression> expressionList = new ArrayDeque<>();

		while (stepIter.hasNext()) {
			assert processIter.hasNext();
			DataAwareStepTypes stepType = stepIter.next();
			ExecutionStep processStep = processIter.next();

			if (processStep != modelMove) {
				// Update the DataState according to the process projection
				for (Entry<String, Object> entry : processStep.entrySet()) {
					Object value = entry.getValue();
					String varName = entry.getKey();					
					DataStateHelper.storeValue(currentDataState, varName.concat("'"), value, value.getClass(), discretizer);
				}
			}

			if (stepType != DataAwareStepTypes.L && processStep.getActivityObject() != null) {
				Expression guard = ((PNWDTransition) processStep.getActivityObject()).getGuard();

				if (guard != null) {
					// Now, extract parts of guard that deal with our variables.
					// For example if X is our variable, then we extract all guards concerning X or X'
					Expression extractedExpression = extractExpression(guard, currentDataState, writeOps);
					if (extractedExpression != null) {
						debug("Extracted Expression: {0} -> {1}", guard, extractedExpression);
						expressionList.add(extractedExpression);
					} else {
						debug("Ignoring Tautology: {0} for Alignment {1} with DataState {2}", guard, alignment,
								currentDataState);
					}
				}
			}

			if (processStep == modelMove) {
				// Know we are interested in the non-prime variables
				writeOps.clear();
				for (String varName : modelMove.keySet()) {
					writeOps.add(varName);
				}
				debug("Current Variables {0}", writeOps);
			}
						
			if (processStep != modelMove) {
				
				currentDataState.movePrimes();
				
				// If a variable in writeOps is written, then we don't care about the value from now on.
				// We already handled the case X==X' in the part before 
				boolean isChanged = writeOps.removeAll(processStep.keySet());

				if (isChanged) {
					debug("Updated Variables {0}", writeOps);
				}
			}
			
		}

		debug("List of Expressions {0}", expressionList);

		return combineExpressions(expressionList);
	}

	private Expression combineExpressions(Deque<Expression> expressionStack) {
		if (!expressionStack.isEmpty()) {
			return doCombineExpressions(expressionStack);
		} else {
			return null;
		}
	}

	private Expression doCombineExpressions(Deque<Expression> expressionStack) {
		if (expressionStack.size() == 1) {
			return expressionStack.pop();
		} else if (!expressionStack.isEmpty()) {
			return new Expression(expressionStack.pop(), "&&", doCombineExpressions(expressionStack));
		} else {
			throw new UnsupportedOperationException();
		}
	}

	private Expression extractExpression(Expression expr, DataState ds, Set<String> writeOps) {
		Object op1 = extractOperand(expr.getOperand1(), ds, writeOps);
		Object op2 = extractOperand(expr.getOperand2(), ds, writeOps);

		if (isConstant(op1) && isConstant(op2)) {
			if (expr.isComparisonExpression()) {
				// It has to be true anyway, we can forget about it
				return null;
			} else {
				// We need to apply the operation
				//TODO we could already calculate this one
				return new Expression(op1, (String) expr.getOperator(), op2);
			}
		} else {
			if (op1 == null && op2 == null) {
				return null;
			} else if (op1 == null && op2 instanceof Expression) {
				return (Expression) op2;
			} else if (op2 == null && op1 instanceof Expression) {
				return (Expression) op1;
			} else {
				return new Expression(op1, (String) expr.getOperator(), op2);
			}			
		}
	}

	private Object extractOperand(Object operand, DataState ds, Set<String> writeOps) {
		if (operand instanceof Expression) {
			return extractExpression((Expression) operand, ds, writeOps);
		} else if (operand instanceof String) {
			String s = (String) operand;
			if (isLiteral(s) || s.equalsIgnoreCase("null")) {
				// Copy as it is
				return s;
			} else {
				Object r = operand;
				if (!writeOps.contains(r)) {
					// It is a variable that was not written by the model move
					// We should be able to to get the value from the DataState
					Class<?> type = varType.get(stripPrime(s));
					if (type == String.class) {
						long tryValue = ds.tryGetLong(s);
						if (tryValue != Long.MIN_VALUE) {
							assert tryValue < Integer.MAX_VALUE : "Invalid value for STRING variable " + s;
							r = "\"" + DataStateHelper.unDiscretize(s, (int) tryValue, discretizer) + "\"";
						}
					} else if (type == Long.class || type == Integer.class) {
						long tryValue = ds.tryGetLong(s);
						if (tryValue != Long.MIN_VALUE) {
							r = Long.valueOf(tryValue);
						}
					} else if (type == Date.class) {
						// Why do we compare dates in this crappy way
						double tryValue = ds.tryGetDouble(s);
						if (!Double.isNaN(tryValue)) {
							r = NumericValueConversion.fromNumericValue(tryValue, type);
						}
					} else if (type == Boolean.class) {
						long tryValue = ds.tryGetLong(s);
						if (tryValue != Long.MIN_VALUE) {
							r = tryValue == 0 ? Boolean.FALSE : Boolean.TRUE;
						}
					} else {
						double tryValue = ds.tryGetDouble(s);
						if (!Double.isNaN(tryValue)) {
							r = Double.valueOf(tryValue);
						}
					}
				}
				return r;
			}
		} else {
			// Constant
			return operand;
		}
	}

	private static boolean isConstant(Object operand) {
		if (operand instanceof String) {
			String s = (String) operand;
			if (isLiteral(s) || s.equalsIgnoreCase("null")) {
				return true;
			} else {
				return false;
			}
		}
		return !(operand instanceof Expression);
	}

	private static boolean isLiteral(Object s) {
		return s instanceof String && ((String) s).charAt(0) == '"';
	}

	private static String stripPrime(String varName) {
		if (varName.charAt(varName.length() - 1) == '\'') {
			return varName.substring(0, varName.length() - 1);
		} else {
			return varName;
		}
	}

	private static void debug(String msg, Object... args) {
		if (IS_DEBUG) {
			if (args != null) {
				System.out.println(MessageFormat.format(msg, args));
			} else {
				System.out.println(msg);
			}
		}
	}

	private static void info(String msg, Object... args) {
		if (args != null) {
			System.out.println(MessageFormat.format(msg, args));
		} else {
			System.out.println(msg);
		}
	}

	public boolean checkConstraints(int modelMoveIndex, EventEntry eventCandidate) {
		DataCheck dataCheck = dataCheckMap.get(modelMoveIndex);
		assert dataCheck != null : "Missing DataCheck for model move at index " + modelMoveIndex;
		return dataCheck.check(modelMoveIndex, eventCandidate);
	}

}
