package org.processmining.logenhancement.merge.alignmentbased.timeconstraints;

import java.util.Collection;
import java.util.Date;
import java.util.ListIterator;
import java.util.Map;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.graphbased.directed.utils.GraphIterator;
import org.processmining.models.graphbased.directed.utils.GraphIterator.EdgeAcceptor;
import org.processmining.models.graphbased.directed.utils.GraphIterator.NodeAcceptor;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

import com.google.common.collect.Maps;

public class TimeConstraintsCheckerStructureImpl implements TimeConstraintsChecker {

	private static final class Move {

		final int index;
		final Transition transition;

		public Move(Transition transition, int index) {
			this.transition = transition;
			this.index = index;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#hashCode()
		 */
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + index;
			// use identify instead of equals, as we always deal with the same set of ExecutionStep objects
			result = prime * result + ((transition == null) ? 0 : System.identityHashCode(transition));
			return result;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Move other = (Move) obj;
			if (index != other.index)
				return false;
			if (transition == null) {
				if (other.transition != null)
					return false;
				// use identify instead of equals, as we always deal with the same set of transition objects
			} else if (!(transition == other.transition))
				return false;
			return true;
		}

		public String toString() {
			return String.format("%s (%s)", transition, index);
		}
	}

	private final Map<Move, Long> earliestTimeMap;
	private final Map<Move, Long> latestTimeMap;

	public TimeConstraintsCheckerStructureImpl(PetriNetWithData net, Marking initialMarking,
			TransEvClassMapping activityMapping, XTrace masterTrace, Alignment alignment, long startDateLimit,
			long endDateLimit) {
		super();
		this.earliestTimeMap = computeEarliestTime(masterTrace, alignment, startDateLimit);
		this.latestTimeMap = computeLatestTime(earliestTimeMap, alignment, masterTrace.get(masterTrace.size() - 1),
				endDateLimit);
	}

	private Map<Move, Long> computeLatestTime(Map<Move, Long> earliestTimeMap, Alignment alignment, XEvent lastEvent,
			long endDateLimit) {
		int size = alignment.getProcessTrace().size();

		Map<Move, Long> latestTimeMap = Maps.newHashMapWithExpectedSize(size);
		Map<Transition, Long> transitionTimeMap = Maps.newHashMapWithExpectedSize(size);

		ListIterator<ExecutionStep> processIter = alignment.getProcessTrace().listIterator(size);
		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator(size);

		while (processIter.hasPrevious()) {
			ExecutionStep processStep = processIter.previous();
			DataAwareStepTypes stepType = stepIter.previous();
			int alignmentIndex = processIter.nextIndex();

			if (stepType != DataAwareStepTypes.L) {
				if (stepType == DataAwareStepTypes.LMGOOD || stepType == DataAwareStepTypes.LMNOGOOD) {
					Long time = earliestTimeMap.get(createMove(alignmentIndex, processStep));
					transitionTimeMap.put(getTransition(processStep), time);
					latestTimeMap.put(new Move(getTransition(processStep), alignmentIndex), time);
				} else {
					// Get the latest time
					long latestTime = Long.MAX_VALUE;
					Collection<PetrinetNode> successors = getSuccessors(getTransition(processStep));
					if (successors.isEmpty()) {
						latestTime = XTimeExtension.instance().extractTimestamp(lastEvent).getTime() + endDateLimit;
					} else {
						for (PetrinetNode node : successors) {
							Long time = transitionTimeMap.get(node);
							if (time != null) {
								latestTime = Math.min(latestTime, time.longValue());
							}
						}
					}
					transitionTimeMap.put(getTransition(processStep), latestTime);
					latestTimeMap.put(new Move(getTransition(processStep), alignmentIndex), latestTime);
				}
			}
		}

		return latestTimeMap;
	}

	private Map<Move, Long> computeEarliestTime(XTrace masterTrace, Alignment alignment, long startDateLimit) {
		int size = alignment.getStepTypes().size();

		Map<Move, Long> earliestTimeMap = Maps.newHashMapWithExpectedSize(size);
		Map<Transition, Long> transitionTimeMap = Maps.newHashMapWithExpectedSize(size);

		ListIterator<ExecutionStep> processIter = alignment.getProcessTrace().listIterator();
		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator();
		ListIterator<XEvent> eventIter = masterTrace.listIterator();

		XEvent lastGoodEvent;
		Date lastSeenTime = new Date(XTimeExtension.instance().extractTimestamp(masterTrace.get(0)).getTime()
				- startDateLimit);

		while (processIter.hasNext()) {
			ExecutionStep processStep = processIter.next();
			DataAwareStepTypes stepType = stepIter.next();
			int alignmentIndex = processIter.previousIndex();

			if (stepType == DataAwareStepTypes.L) {
				eventIter.next();
			} else if (stepType == DataAwareStepTypes.LMGOOD || stepType == DataAwareStepTypes.LMNOGOOD) {
				lastGoodEvent = eventIter.next();
				lastSeenTime = XTimeExtension.instance().extractTimestamp(lastGoodEvent);
			} else {
				// Do not move in Events
			}

			if (stepType != DataAwareStepTypes.L) {
				if (stepType == DataAwareStepTypes.LMGOOD || stepType == DataAwareStepTypes.LMNOGOOD) {
					transitionTimeMap.put(getTransition(processStep), lastSeenTime.getTime());
					earliestTimeMap.put(new Move(getTransition(processStep), alignmentIndex), lastSeenTime.getTime());
				} else {
					// Get the earliest time
					long earliestTime = Long.MIN_VALUE;
					Collection<PetrinetNode> predecessors = getPredecessors(getTransition(processStep));
					if (predecessors.isEmpty()) {
						earliestTime = XTimeExtension.instance().extractTimestamp(masterTrace.get(0)).getTime()
								- startDateLimit;
					} else {
						for (PetrinetNode node : predecessors) {
							Long time = transitionTimeMap.get(node);
							if (time != null) {
								earliestTime = Math.max(earliestTime, time.longValue());
							}
						}
					}
					transitionTimeMap.put(getTransition(processStep), earliestTime);
					earliestTimeMap.put(new Move(getTransition(processStep), alignmentIndex), earliestTime);
				}
			}
		}

		return earliestTimeMap;
	}

	private Collection<PetrinetNode> getSuccessors(Transition transition) {
		final NodeAcceptor<PetrinetNode> nodeAcceptor = new NodeAcceptor<PetrinetNode>() {
			public boolean acceptNode(PetrinetNode node, int depth) {
				return ((depth != 0) && (node instanceof Transition));
			}
		};
		Collection<PetrinetNode> transitions = GraphIterator.getDepthFirstSuccessors(transition, transition.getGraph(),
				new EdgeAcceptor<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>>() {

					public boolean acceptEdge(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge,
							int depth) {
						return !nodeAcceptor.acceptNode(edge.getSource(), depth);
					}
				}, nodeAcceptor);
		return transitions;
	}

	private Collection<PetrinetNode> getPredecessors(Transition transition) {
		final NodeAcceptor<PetrinetNode> nodeAcceptor = new NodeAcceptor<PetrinetNode>() {
			public boolean acceptNode(PetrinetNode node, int depth) {
				return ((depth != 0) && (node instanceof Transition));
			}
		};
		Collection<PetrinetNode> transitions = GraphIterator.getDepthFirstPredecessors(transition,
				transition.getGraph(),
				new EdgeAcceptor<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>>() {

					public boolean acceptEdge(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge,
							int depth) {
						return !nodeAcceptor.acceptNode(edge.getTarget(), depth);
					}
				}, nodeAcceptor);
		return transitions;
	}

	private Transition getTransition(ExecutionStep processStep) {
		return (Transition) processStep.getActivityObject();
	}

	public TimeConstraints getTimeConstraints(int alignmentIndex, ExecutionStep modelMove) {
		TimeConstraints time = new TimeConstraints();
		time.setTimeAfter(new Date(latestTimeMap.get(createMove(alignmentIndex, modelMove))));
		time.setTimeBefore(new Date(earliestTimeMap.get(createMove(alignmentIndex, modelMove))));
		assert time.getTimeAfter().after(time.getTimeBefore()) || time.getTimeAfter().equals(time.getTimeBefore()) : "Invalid Time Constraint, timeAfter is before timeBefore "
				+ time + " for move " + modelMove;
		return time;
	}

	private Move createMove(int alignmentIndex, ExecutionStep modelMove) {
		return new Move((Transition) modelMove.getActivityObject(), alignmentIndex);
	}

}
