package org.processmining.logenhancement.merge.alignmentbased.timeconstraints;

import java.text.MessageFormat;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.ListIterator;
import java.util.Map;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.semantics.IllegalTransitionException;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.PetrinetSemantics;
import org.processmining.models.semantics.petrinet.impl.PetrinetSemanticsFactory;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

import com.google.common.collect.ImmutableMap;

public class TimeConstraintsCheckerTokenImpl implements TimeConstraintsChecker {

	private static final class ModelMove {

		final int index;
		final Transition transition;

		public ModelMove(Transition transition, int index) {
			this.transition = transition;
			this.index = index;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#hashCode()
		 */
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + index;
			// use identify instead of equals, as we always deal with the same set of transition objects
			result = prime * result + ((transition == null) ? 0 : System.identityHashCode(transition));
			return result;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			ModelMove other = (ModelMove) obj;
			if (index != other.index)
				return false;
			if (transition == null) {
				if (other.transition != null)
					return false;
				// use identify instead of equals, as we always deal with the same set of transition objects
			} else if (!(transition == other.transition))
				return false;
			return true;
		}

		public String toString() {
			return String.format("%s (%s)", transition, index);
		}
	}

	private static final boolean IS_DEBUG = false;

	private long endDateLimit;

	/**
	 * Map between model moves and Time Info sorted by occurrence of the model
	 * move in the Alignment
	 */
	private final ImmutableMap<ModelMove, TimeConstraints> timeMap;

	public TimeConstraintsCheckerTokenImpl(PetriNetWithData net, Marking initialMarking, TransEvClassMapping activityMapping,
			XTrace masterTrace, Alignment alignment, long startDateLimit, long endDateLimit) {

		this.endDateLimit = endDateLimit;
		this.timeMap = initializeTimeMap(alignment);

		PetrinetSemantics petriNetSemantics = PetrinetSemanticsFactory.regularPetrinetSemantics(Petrinet.class);
		petriNetSemantics.initialize(net.getTransitions(), initialMarking);

		ListIterator<ExecutionStep> processIter = alignment.getProcessTrace().listIterator();
		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator();
		ListIterator<XEvent> eventIter = masterTrace.listIterator();
		assert alignment.getProcessTrace().size() == alignment.getStepTypes().size();

		XEvent lastGoodEvent;
		Date lastSeenTime = new Date(XTimeExtension.instance().extractTimestamp(masterTrace.get(0)).getTime()
				- startDateLimit);

		while (processIter.hasNext()) {
			ExecutionStep processStep = processIter.next();
			DataAwareStepTypes stepType = stepIter.next();
			int alignmentIndex = processIter.previousIndex();

			if (stepType == DataAwareStepTypes.L) {
				eventIter.next();
			} else if (stepType == DataAwareStepTypes.LMGOOD || stepType == DataAwareStepTypes.LMNOGOOD) {
				lastGoodEvent = eventIter.next();
				lastSeenTime = XTimeExtension.instance().extractTimestamp(lastGoodEvent);
			} else {
				// Do not move in Events
			}
			int eventIndex = eventIter.previousIndex();

			if (stepType != DataAwareStepTypes.L) {
				// Find corresponding transition
				Transition nextTransition = safeGetTransition(processStep);

				if (stepType == DataAwareStepTypes.MREAL) {
					// Look ahead for the latest time
					Date timeAfter = lookAhead(masterTrace, net, nextTransition, petriNetSemantics.getCurrentState(),
							alignment, alignmentIndex, eventIndex);
					timeMap.get(new ModelMove(nextTransition, alignmentIndex)).setTimeAfter(timeAfter);
				}

				Collection<Transition> executableTransitions = petriNetSemantics.getExecutableTransitions();

				// Look if any of our model moves could be executed now
				for (Transition executableTransition : executableTransitions) {
					for (ModelMove modelMove : timeMap.keySet()) {
						// Ignore already processed model moves
						if (modelMove.index >= alignmentIndex) {
							if (executableTransition == modelMove.transition) { // Identify equal is safe here!
								// We could fire this model move now, we still have so see if we could finish the trace as in the alignment from here
								if (couldFinish(net, modelMove.transition, petriNetSemantics.getCurrentState(),
										alignment, alignmentIndex)) {
									// If we could still finish firing the model move now, then remember the earliest time
									if (timeMap.get(modelMove).getTimeBefore() == null) {
										timeMap.get(modelMove).setTimeBefore(lastSeenTime);
									}
								}
								// Stop at first match, as we only want to record the time for  
								// the next model move in the Alignment for this transition.
								// We _assume_ that timeMap is sorted by appearance in the Alignment.
								break;
							}
						}
					}
				}

				try {
					petriNetSemantics.executeExecutableTransition(nextTransition);
				} catch (IllegalTransitionException e) {
					throw new RuntimeException("Alignment suggests that the transition " + nextTransition
							+ " should be executed next, but it is not exectable! Full Alignment " + alignment, e);
				}

			}
		}

		debug("{0}", timeMap);

	}

	private boolean couldFinish(PetriNetWithData net, Transition modelMove, Marking currentState, Alignment alignment,
			int alignmentIndex) {

		debug("Looking ahead if we could still finish firing {0} now, at marking: ", modelMove);
		debug("{0}", currentState);

		PetrinetSemantics couldFinishSemantics = PetrinetSemanticsFactory.regularPetrinetSemantics(Petrinet.class);
		couldFinishSemantics.initialize(net.getTransitions(), new Marking(currentState));

		// Execute the model move transition
		try {
			couldFinishSemantics.executeExecutableTransition(modelMove);
		} catch (IllegalTransitionException e) {
			throw new RuntimeException("We should really be able to execute the transition " + modelMove
					+ " next, but it is not exectable!", e);
		}

		ListIterator<ExecutionStep> processIter = alignment.getProcessTrace().listIterator(alignmentIndex);
		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator(alignmentIndex);

		while (stepIter.hasNext()) {
			ExecutionStep processStep = processIter.next();
			DataAwareStepTypes stepType = stepIter.next();

			if (stepType != DataAwareStepTypes.L) {
				Transition nextTransition = safeGetTransition(processStep);

				if (nextTransition != modelMove) {
					if (couldFinishSemantics.getExecutableTransitions().contains(nextTransition)) {
						try {
							debug("Firing {0}", nextTransition);
							couldFinishSemantics.executeExecutableTransition(nextTransition);
						} catch (IllegalTransitionException e) {
							throw new RuntimeException("Transition should really be executable " + nextTransition, e);
						}
					} else {
						debug("We could NOT finish anymore, if firing {0} now!", modelMove.getLabel());
						return false;
					}
				}
			}
		}

		debug("We could finish the trace, if firing {0} now!", modelMove.getLabel());
		return true;
	}

	private static void debug(String msg, Object... args) {
		if (IS_DEBUG) {
			if (args != null) {
				System.out.println(MessageFormat.format(msg, args));
			} else {
				System.out.println(msg);
			}
		}
	}

	private Date lookAhead(XTrace masterTrace, PetriNetWithData net, Transition moveInModelTransition,
			Marking currentState, Alignment alignment, int alignmentIndex, int eventIndex) {

		debug("Looking ahead transition {0} starting with marking: ", moveInModelTransition);
		debug("{0}", currentState);

		PetrinetSemantics lookAheadSemantics = PetrinetSemanticsFactory.regularPetrinetSemantics(Petrinet.class);
		lookAheadSemantics.initialize(net.getTransitions(), new Marking(currentState));

		ListIterator<ExecutionStep> processIter = alignment.getProcessTrace().listIterator(alignmentIndex + 1);
		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator(alignmentIndex + 1);
		ListIterator<XEvent> eventIter = masterTrace.listIterator(eventIndex + 1);

		while (stepIter.hasNext()) {
			ExecutionStep processStep = processIter.next();
			DataAwareStepTypes stepType = stepIter.next();

			XEvent currentEvent = null;
			if (stepType == DataAwareStepTypes.L) {
				eventIter.next();
			} else if (stepType == DataAwareStepTypes.LMGOOD || stepType == DataAwareStepTypes.LMNOGOOD) {
				currentEvent = eventIter.next();
			} else {
				// Do not move in Events
			}

			if (stepType != DataAwareStepTypes.L) {
				// Lookup transition in Petri Net
				Transition nextTransition = safeGetTransition(processStep);

				if (lookAheadSemantics.getExecutableTransitions().contains(nextTransition)) {
					// We can go forward in the Alignment without executing the model move
					try {
						debug("Firing {0}", nextTransition);
						lookAheadSemantics.executeExecutableTransition(nextTransition);
					} catch (IllegalTransitionException e) {
						throw new RuntimeException("Transition should really be executable " + nextTransition, e);
					}
				} else {
					if (stepType == DataAwareStepTypes.LMGOOD || stepType == DataAwareStepTypes.LMNOGOOD) {
						debug("Cannot execute Transition {0} - found Move in Both {1}", nextTransition,
								XConceptExtension.instance().extractName(currentEvent));
						return XTimeExtension.instance().extractTimestamp(currentEvent);
					} else {
						debug("Cannot execute Transition {0} - ignoring!", nextTransition);
					}
				}

				if (!lookAheadSemantics.getExecutableTransitions().contains(moveInModelTransition)) {
					// We cannot execute the model move transition anymore,
					// so the last "move in both" represents the latest possible execution time
					debug("Cannot execute {0} anymore!", moveInModelTransition);
					if (stepType == DataAwareStepTypes.LMGOOD || stepType == DataAwareStepTypes.LMNOGOOD) {
						return XTimeExtension.instance().extractTimestamp(currentEvent);
					} else {
						// Last move was not linked an event, so just go through the alignment and get the next move in model
						debug("No move in both in between - giving up!");
						break;
					}
				} else {
					debug("Transition {0} still executable", moveInModelTransition);
				}
			}
		}

		// If we stopped in the middle, look for the next move in both
		while (stepIter.hasNext()) {
			debug("Looking for next move in both ...");
			DataAwareStepTypes stepType = stepIter.next();

			if (stepType == DataAwareStepTypes.L) {
				eventIter.next();
			} else if (stepType == DataAwareStepTypes.LMGOOD || stepType == DataAwareStepTypes.LMNOGOOD) {
				return XTimeExtension.instance().extractTimestamp(eventIter.next());
			}
		}
		debug("No limit!");
		// If everything fails, default limit after last observed event
		return new Date(XTimeExtension.instance().extractTimestamp(masterTrace.get(masterTrace.size() - 1)).getTime()
				+ endDateLimit);
	}

	private ImmutableMap<ModelMove, TimeConstraints> initializeTimeMap(Alignment alignment) {

		// Sort by insertion order!
		Map<ModelMove, TimeConstraints> timeMap = new LinkedHashMap<ModelMove, TimeConstraints>();

		ListIterator<ExecutionStep> processIter = alignment.getProcessTrace().listIterator();
		ListIterator<DataAwareStepTypes> stepIter = alignment.getStepTypes().listIterator();

		assert alignment.getProcessTrace().size() == alignment.getStepTypes().size();

		while (processIter.hasNext()) {
			ExecutionStep processStep = processIter.next();
			DataAwareStepTypes stepType = stepIter.next();

			if (stepType == DataAwareStepTypes.MREAL) {
				Transition transition = safeGetTransition(processStep);
				timeMap.put(new ModelMove(transition, processIter.previousIndex()), new TimeConstraints());
			}
		}

		return ImmutableMap.copyOf(timeMap);
	}

	private static Transition safeGetTransition(ExecutionStep modelMove) {
		assert modelMove.getActivityObject() != null && modelMove.getActivityObject() instanceof Transition : "Missing Transition for ExecutionStep "
				+ modelMove;
		return (Transition) modelMove.getActivityObject();
	}

	/* (non-Javadoc)
	 * @see org.processmining.plugins.fm.project4.ITimeConstraintsChecker#getTimeConstraints(int, org.processmining.plugins.DataConformance.framework.ExecutionStep)
	 */
	@Override
	public TimeConstraints getTimeConstraints(int alignmentIndex, ExecutionStep modelMove) {
		return timeMap.get(new ModelMove(safeGetTransition(modelMove), alignmentIndex));
	}

}
