package org.processmining.logenhancement.transforming;

import org.deckfour.xes.classification.XEventAndClassifier;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventLifeTransClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;

/**
 * Plug-in that transforms a log at attribute level.
 * 
 * @author Eduardo Gonzalez Lopez de Murillas
 * 
 */
@Plugin(name = "Add/Change concept:name of event from classifier", returnLabels = {
		"Log with changed 'concept:name'" }, returnTypes = { XLog.class }, parameterLabels = { "Log",
				"Concept Name Parameters", "Activity Mapping Parameters" }, userAccessible = true, categories = {
						PluginCategory.Enhancement }, help = "Adds an attribute 'concept:name' to each event with the corresponding activity name from the selected classifier.")
public class AddConceptNamePlugin {

	private static class AddConceptNameParameters {

		/**
		 * Classifier parameter. This determines which classifier will be used
		 * during the transformation.
		 */
		private XEventClassifier classifier;

		/**
		 * Create default parameter values.
		 */
		public AddConceptNameParameters() {
			classifier = new XEventAndClassifier(new XEventNameClassifier(), new XEventLifeTransClassifier());
		}

		/**
		 * Set the classifier to the given classifier.
		 * 
		 * @param classifier
		 *            The given classifier.
		 */
		public void setClassifier(XEventClassifier classifier) {
			if (classifier != null) {
				this.classifier = classifier;
			}
		}

		/**
		 * Gets the classifier.
		 * 
		 * @return The classifier.
		 */
		public XEventClassifier getClassifier() {
			return classifier;
		}

		/**
		 * Returns whether these parameter values are equal to the given
		 * parameter values.
		 * 
		 * @param object
		 *            The given parameter values.
		 * @return Whether these parameter values are equal to the given
		 *         parameter values.
		 */
		public boolean equals(Object object) {
			if (object instanceof AddConceptNameParameters) {
				AddConceptNameParameters parameters = (AddConceptNameParameters) object;
				if (classifier.equals(parameters.classifier)) {
					return true;
				}
			}
			return false;
		}

		/**
		 * Returns the hash code for these parameters.
		 */
		public int hashCode() {
			return classifier.hashCode();
		}
	}

	/**
	 * Transform using provided parameter values.
	 * 
	 * @param context
	 *            The given plug-in context.
	 * @param log
	 *            The given event log.
	 * @param parameters
	 *            The given parameters to select the right classifier
	 * @return The transformed event log with added concept names.
	 */
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "Eduardo González López de Murillas", email = "e.gonzalez@tue.nl")
	@PluginVariant(variantLabel = "Add Concept Name to Events from Classifier", requiredParameterLabels = { 0, 1 })
	public XLog addConceptNameDefault(PluginContext context, XLog log, AddConceptNameParameters parameters) {
		return addConceptName(context, log, parameters);
	}

	/**
	 * Transform using provided parameter values.
	 * 
	 * @param context
	 *            The given GUI-aware plug-in context.
	 * @param log
	 *            The given event log.
	 * @return The transformed event log with added concept names.
	 */
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "Eduardo González López de Murillas", email = "e.gonzalez@tue.nl")
	@PluginVariant(variantLabel = "Add Concept Name to Events from Classifier", requiredParameterLabels = { 0 })
	public XLog addConceptNameDefault(UIPluginContext context, XLog log) {
		AddConceptNameParameters parameters = new AddConceptNameParameters();

		XEventClassifier ec = null;
		try {
			ec = (XEventClassifier) ProMUIHelper.queryForObject(context, "Select a classifier",
					log.getClassifiers().toArray());
			if (ec != null) {
				parameters.setClassifier(ec);
			}
		} catch (UserCancelledException e) {
			return null;
		}

		return addConceptName(context, log, parameters);
	}

	/*
	 * The actual transformation an event log given parameter values.
	 */
	private XLog addConceptName(PluginContext context, XLog log, AddConceptNameParameters parameters) {

		XLog newLog = (XLog) log.clone();

		XEventClassifier classifier = parameters.getClassifier();

		/*
		 * Inform the progress bar when we're done.
		 */
		context.getProgress().setMaximum(newLog.size());

		for (XTrace trace : newLog) {
			for (XEvent event : trace) {

				String identity = classifier.getClassIdentity(event);

				event.getAttributes().put(XConceptExtension.KEY_NAME,
						new XAttributeLiteralImpl(XConceptExtension.KEY_NAME, identity));
			}

			context.getProgress().inc();
		}

		/*
		 * Return the model.
		 */
		return newLog;
	}

}