package org.processmining.logenhancement.transforming;

import java.util.HashSet;
import java.util.Set;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;

public class ConcatenateEventLogs {

	@Plugin(name = "Concatenate/Union of event logs ", // 
			returnLabels = { " Concatenated/Union Log" }, // 
			returnTypes = { XLog.class }, userAccessible = true, //
			categories = { PluginCategory.Enhancement }, // 
			parameterLabels = { "Event Logs" }, mostSignificantResult = 1, //
			help = "Concatenates two or more event logs. The resulting log contains the union of traces from both event logs. "
					+ "Traces are added to the resulting event log as-is, no merging takes place.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog concatenateEventsLogs(PluginContext context, XLog... logs) throws Exception {

		context.getProgress().setCaption("Concatenating logs ...");

		return doConcatenateEventLogs(logs);
	}

	public XLog doConcatenateEventLogs(XLog... logs) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();

		Set<XAttribute> globalEventAttributes = new HashSet<>();
		Set<XAttribute> globalTraceAttributes = new HashSet<>();
		Set<XEventClassifier> classifiers = new HashSet<>();

		XLog concatenatedLog = factory.createLog();
		for (XLog log : logs) {
			concatenatedLog.getAttributes().putAll((XAttributeMap) log.getAttributes().clone());

			concatenatedLog.getExtensions().addAll(log.getExtensions());
			globalEventAttributes.addAll(log.getGlobalEventAttributes());
			globalTraceAttributes.addAll(log.getGlobalTraceAttributes());
			classifiers.addAll(log.getClassifiers());

			for (XTrace t : log) {
				concatenatedLog.add((XTrace) t.clone());
			}
		}

		concatenatedLog.getGlobalEventAttributes().addAll(globalEventAttributes);
		concatenatedLog.getGlobalTraceAttributes().addAll(globalTraceAttributes);
		concatenatedLog.getClassifiers().addAll(classifiers);

		return concatenatedLog;
	}

	@Plugin(name = "Concatenate/Union of event logs (without copy)", // 
			returnLabels = { " Concatenated/Union Log" }, // 
			returnTypes = { XLog.class }, userAccessible = true, //
			categories = { PluginCategory.Enhancement }, // 
			parameterLabels = { "Event Logs" }, mostSignificantResult = 1, //
			help = "Concatenates two or more event logs. The resulting log contains the union of traces from both event logs. "
					+ "Traces are added to the resulting event log as-is, no merging takes place."
					+ " This plug-in variant does not copy the individual traces. It merly creates a wrapper log that contains all the original traces. This reduces the memory requirements.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog concatenateEventsLogsWithoutCopy(PluginContext context, XLog... logs) throws Exception {

		context.getProgress().setCaption("Concatenating logs ...");

		XFactory factory = XFactoryRegistry.instance().currentDefault();

		Set<XAttribute> globalEventAttributes = new HashSet<>();
		Set<XAttribute> globalTraceAttributes = new HashSet<>();
		Set<XEventClassifier> classifiers = new HashSet<>();

		XLog concatenatedLog = factory.createLog();
		for (XLog log : logs) {
			concatenatedLog.getAttributes().putAll(log.getAttributes());
			concatenatedLog.getExtensions().addAll(log.getExtensions());
			globalEventAttributes.addAll(log.getGlobalEventAttributes());
			globalTraceAttributes.addAll(log.getGlobalTraceAttributes());
			classifiers.addAll(log.getClassifiers());

			for (XTrace t : log) {
				concatenatedLog.add(t);
			}
		}

		concatenatedLog.getGlobalEventAttributes().addAll(globalEventAttributes);
		concatenatedLog.getGlobalTraceAttributes().addAll(globalTraceAttributes);
		concatenatedLog.getClassifiers().addAll(classifiers);

		return concatenatedLog;
	}

}
