package org.processmining.logenhancement.transforming;

import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension.StandardModel;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

import com.google.common.collect.Sets;

public class CreateEventFromAttribute {

	@Plugin(name = "Create new events based on timestamp attributes", parameterLabels = { "Event Log" }, //
	returnLabels = { "Enhanced Log" }, returnTypes = { XLog.class }, userAccessible = true, categories = { PluginCategory.Enhancement },//
	help = "Creates a new event based on the timestamp information of an attribute in the event log. Only supports the 'concept:name' classifier.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog createEventFromTimestamps(UIPluginContext context, XLog log) throws UserCancelledException {

		context.getProgress().setMaximum(log.size());

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		if (!(classifier.getDefiningAttributeKeys().length == 1 && classifier.getDefiningAttributeKeys()[0]
				.equals("concept:name"))) {
			throw new RuntimeException("This plug-in does only work with the XEventNameClassifier (concept:name)");
		}
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, classifier);
		Set<XEventClass> eventNames = Sets.newTreeSet(logInfo.getEventClasses().getClasses());

		List<XEventClass> selectedEventClasses = ProMUIHelper.queryForObjects(context,
				"Please choose event classes to be searched for timestamp attributes", eventNames);
		Boolean traceIncluded = ProMUIHelper.queryForObject(context, "Do you want to include trace attributes?",
				new Boolean[] { true, false });

		Set<String> attributeSet = new TreeSet<>();
		Set<String> timestampAttributes = new TreeSet<>();

		for (XTrace t : log) {
			if (traceIncluded) {
				for (XAttribute a : t.getAttributes().values()) {
					if (a instanceof XAttributeTimestamp) {
						timestampAttributes.add(a.getKey());
					}
					attributeSet.add(a.getKey());
				}
			}
			for (XEvent e : t) {
				if (selectedEventClasses.contains(logInfo.getEventClasses().getClassOf(e))) {
					for (XAttribute a : e.getAttributes().values()) {
						if (a instanceof XAttributeTimestamp) {
							timestampAttributes.add(a.getKey());
						}
						attributeSet.add(a.getKey());
					}
				}
			}
		}

		String attributeName = ProMUIHelper.queryForObject(context, "Please choose which timestamp attribute to use",
				timestampAttributes);
		String newEventName = ProMUIHelper.queryForString(context,
				"Please provide a concept:name for the events to be created");
		List<String> attributesToCopy = ProMUIHelper.queryForStrings(context,
				"Please select which other attributes of the source event to copy", attributeSet);

		XFactory factory = XFactoryRegistry.instance().currentDefault();

		return doCreateEventFromTimestamp(log, logInfo, factory, selectedEventClasses, traceIncluded, attributeName,
				newEventName, attributesToCopy);
	}

	public XLog doCreateEventFromTimestamp(XLog log, XLogInfo logInfo, XFactory factory,
			List<XEventClass> selectedEventClasses, Boolean traceIncluded, String attributeName, String newEventName,
			List<String> attributesToCopy) {

		XLog newLog = factory.createLog((XAttributeMap) log.getAttributes().clone());

		for (XTrace t : log) {
			XTrace newTrace = factory.createTrace((XAttributeMap) t.getAttributes().clone());

			if (traceIncluded) {
				createAndAddEvent(attributeName, newEventName, t, newTrace, factory, attributesToCopy);
			}

			for (XEvent e : t) {
				newTrace.insertOrdered((XEvent) e.clone());

				if (selectedEventClasses.contains(logInfo.getEventClasses().getClassOf(e))) {
					createAndAddEvent(attributeName, newEventName, e, newTrace, factory, attributesToCopy);
				}
			}

			newLog.add(newTrace);
		}

		return newLog;
	}

	private void createAndAddEvent(String attributeName, String newEventName, XAttributable attributable,
			XTrace newTrace, XFactory factory, List<String> attributesToCopy) {
		XAttribute attribute = attributable.getAttributes().get(attributeName);
		if (attribute != null && attribute instanceof XAttributeTimestamp) {
			XEvent event = createEventForAttribute(factory, (XAttributeTimestamp) attribute, newEventName);			
			if (!attributable.getAttributes().isEmpty()) {
				for (String key : attributesToCopy) {
					XAttribute oldAttributes = attributable.getAttributes().get(key);
					if (oldAttributes != null) {
						event.getAttributes().put(oldAttributes.getKey(), (XAttribute) oldAttributes.clone());
					}
				}
			}
			if (event != null) {
				newTrace.insertOrdered(event);
			}
		}
	}

	private XEvent createEventForAttribute(XFactory factory, XAttributeTimestamp attribute, String newEventName) {
		XEvent newEvent = factory.createEvent();
		XTimeExtension.instance().assignTimestamp(newEvent, attribute.getValueMillis());
		XConceptExtension.instance().assignName(newEvent, newEventName);
		XLifecycleExtension.instance().assignStandardTransition(newEvent, StandardModel.COMPLETE);
		return newEvent;
	}
}
