package org.processmining.logenhancement.transforming;

import java.awt.BorderLayout;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceListener;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import javax.swing.DefaultListModel;
import javax.swing.DropMode;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.TransferHandler;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.ProMList;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

import com.google.common.base.Function;
import com.google.common.collect.Ordering;
import com.google.common.collect.Sets;

public class FixOrderingSameTimePlugin {

	static class OrderingChoicePanel<T> extends JPanel {

		private static class DragListenerImpl<T> implements DragSourceListener, DragGestureListener {

			private final JList<T> list;
			private final DragSource ds = new DragSource();

			public DragListenerImpl(JList<T> list) {
				this.list = list;
				ds.createDefaultDragGestureRecognizer(list, DnDConstants.ACTION_MOVE, this);
			}

			public void dragGestureRecognized(DragGestureEvent dge) {
				StringSelection transferable = new StringSelection(Integer.toString(list.getSelectedIndex()));
				ds.startDrag(dge, DragSource.DefaultCopyDrop, transferable, this);
			}

			public void dragEnter(DragSourceDragEvent dsde) {
			}

			public void dragExit(DragSourceEvent dse) {
			}

			public void dragOver(DragSourceDragEvent dsde) {
			}

			public void dragDropEnd(DragSourceDropEvent dsde) {
			}

			public void dropActionChanged(DragSourceDragEvent dsde) {
			}
		}

		private static class DropHandlerImpl<T> extends TransferHandler {

			private static final long serialVersionUID = -3468373344687124791L;

			private final DefaultListModel<T> listModel;

			public DropHandlerImpl(DefaultListModel<T> listModel) {
				this.listModel = listModel;
			}

			public boolean canImport(TransferHandler.TransferSupport support) {
				if (!support.isDataFlavorSupported(DataFlavor.stringFlavor)) {
					return false;
				}
				JList.DropLocation dl = (JList.DropLocation) support.getDropLocation();
				if (dl.getIndex() == -1) {
					return false;
				} else {
					return true;
				}
			}

			public boolean importData(TransferHandler.TransferSupport support) {
				if (!canImport(support)) {
					return false;
				}

				Transferable transferable = support.getTransferable();
				String indexString;
				try {
					indexString = (String) transferable.getTransferData(DataFlavor.stringFlavor);
				} catch (Exception e) {
					return false;
				}

				int sourceIndex = Integer.parseInt(indexString);
				JList.DropLocation dl = (JList.DropLocation) support.getDropLocation();
				int dropTargetIndex = dl.getIndex();
				T element = listModel.remove(sourceIndex);
				if (sourceIndex < dropTargetIndex) {
					listModel.insertElementAt(element, dropTargetIndex - 1);
				} else {
					listModel.insertElementAt(element, dropTargetIndex);
				}
				return true;
			}
		}

		private static final long serialVersionUID = -6547392010448275699L;

		private final ProMList<T> choiceField;
		private final DefaultListModel<T> listModel;

		@SuppressWarnings("unchecked")
		public OrderingChoicePanel(String title, Iterable<T> choices) {
			super(new BorderLayout());
			listModel = new DefaultListModel<>();
			for (T c : choices) {
				listModel.addElement(c);

			}
			choiceField = new ProMList<>(title, listModel);
			choiceField.getList().setDragEnabled(true);
			choiceField.getList().setDropMode(DropMode.INSERT);
			choiceField.getList().setTransferHandler(new DropHandlerImpl<>(listModel));
			new DragListenerImpl<>(choiceField.getList());

			choiceField.setPreferredSize(null);
			add(choiceField, BorderLayout.CENTER);
		}

		public InteractionResult getUserChoice(UIPluginContext context, String query) {
			return context.showConfiguration(query, this);
		}

		@SuppressWarnings("unchecked")
		public List<T> getResult() {
			return (List<T>) Arrays.asList(listModel.toArray());
		}

	}

	@Plugin(name = "Change order of events with same time:timestamp", parameterLabels = { "Event Log" }, //
			categories = { PluginCategory.Enhancement },
	returnLabels = { "Enhanced Log" }, returnTypes = { XLog.class }, userAccessible = true,// 
	help = "Re-orders events with the exact same time:timestamp based on a explicit secondary ordering of event classes.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog fixOrderingSameTime(UIPluginContext context, XLog log) throws UserCancelledException {

		context.getProgress().setMaximum(log.size());

		final XEventClasses eventClasses = queryEventClasses(context, log);
		final Ordering<XEvent> compundOrdering = createOrdering(context, log, eventClasses);

		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog newLog = factory.createLog((XAttributeMap) log.getAttributes().clone());

		return doFixOrderingSameTime(context, log, compundOrdering, newLog);
	}

	public XLog doFixOrderingSameTime(PluginContext context, XLog log, final Ordering<XEvent> compundOrdering,
			XLog newLog) {
		for (XTrace t : log) {
			XTrace newTrace = (XTrace) t.clone();
			Collections.sort(newTrace, compundOrdering);
			newLog.add(newTrace);
			context.getProgress().inc();
		}

		return newLog;
	}

	@Plugin(name = "Change order of events with same time:timestamp (In place)", parameterLabels = { "Event Log" }, //
	returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1,//
	help = "Re-orders events with the exact same 'time:timestamp' based on a explicit ordering of event classes. Does NOT create a new event log, but changes the input.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void fixOrderingSameTimInPlace(UIPluginContext context, XLog log) throws UserCancelledException {

		context.getProgress().setMaximum(log.size());

		final XEventClasses eventClasses = queryEventClasses(context, log);
		final Ordering<XEvent> compundOrdering = createOrdering(context, log, eventClasses);

		doFixOrderingSameTimeInPlace(context.getProgress(), log, compundOrdering);

	}

	public void doFixOrderingSameTimeInPlace(Progress progress, XLog log, final Ordering<XEvent> manualOrdering) {
		for (XTrace t : log) {
			Collections.sort(t, manualOrdering);
			progress.inc();
		}
	}

	private static XEventClasses queryEventClasses(UIPluginContext context, XLog log) throws UserCancelledException {
		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, classifier);
		final XEventClasses eventClasses = logInfo.getEventClasses();
		return eventClasses;
	}

	private static Ordering<XEvent> createOrdering(UIPluginContext context, XLog log, final XEventClasses eventClasses)
			throws UserCancelledException {
		List<XEventClass> orderedEventClasses = queryUserDefinedOrdering(context, log, eventClasses);
		Ordering<XEvent> timeOrdering = new Ordering<XEvent>() {

			public int compare(XEvent o1, XEvent o2) {
				final XTimeExtension time = XTimeExtension.instance();
				return time.extractTimestamp(o1).compareTo(time.extractTimestamp(o2));
			}
		};
		Ordering<XEventClass> eventClassOrdering = Ordering.explicit(orderedEventClasses);

		final Ordering<XEvent> compundOrdering = timeOrdering.compound(eventClassOrdering
				.onResultOf(new Function<XEvent, XEventClass>() {

					public XEventClass apply(XEvent event) {
						return eventClasses.getClassOf(event);
					}
				}));
		return compundOrdering;
	}

	private static List<XEventClass> queryUserDefinedOrdering(UIPluginContext context, XLog log,
			XEventClasses eventClasses) throws UserCancelledException {
		Set<XEventClass> eventNames = Sets.newTreeSet(eventClasses.getClasses());
		return getEventOrdering(context, eventNames);
	}

	private static <T> List<T> getEventOrdering(UIPluginContext context, Set<T> eventNames)
			throws UserCancelledException {
		OrderingChoicePanel<T> orderingChoicePanel = new OrderingChoicePanel<>(null, eventNames);
		InteractionResult choice = orderingChoicePanel.getUserChoice(context,
				"Please re-order event classes as they should be ordered in the event log");

		if (choice == InteractionResult.FINISHED || choice == InteractionResult.CONTINUE) {
			return orderingChoicePanel.getResult();
		} else {
			throw new UserCancelledException();
		}
	}

}