package org.processmining.logenhancement.transforming;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XOrganizationalExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;

import com.google.common.collect.ImmutableSet;

/**
 * Plug-in that moves all common attributes (i.e. appear in each event with the
 * same value) to trace level. PLEASE NOTE: This filter will update the original
 * XLog instead of creating a new XLog.
 * 
 * @author F. Mannhardt
 * 
 */
public class MoveCommonAttributesToTracePlugin {

	private final static class ValueCounter {
		private int counter = 1;
		private XAttribute value;

		public ValueCounter(XAttribute value) {
			super();
			this.value = value;
		}

		public void inc(XAttribute value) {
			if (this.value.equals(value)) {
				counter++;
			}
		}

		public int getCounter() {
			return counter;
		}

		public XAttribute getValue() {
			return value;
		}
	}
	
	@Plugin(name = "Copy attributes from event-level to trace-level", parameterLabels = { "Event Log" }, //			
			returnLabels = {}, returnTypes = {}, categories = {
					PluginCategory.Enhancement }, level = PluginLevel.PeerReviewed, userAccessible = true, mostSignificantResult = -1, 
							help = "Copy the selected attributes from event-level to trace-level. "
									+ "When multiple events have different values for an attribute to be moved at trace level,"
									+ "the latest value is chosen. If no event in the trace contains any of the chosen attributes,"
									+ "these attributes are not added at trace level")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " M. de Leoni", email = "m.d.leoni@tue.nl")
	public void moveAttributesToTraceProMLite(UIPluginContext context, XLog log) throws UserCancelledException {


		List<String> attributeKeys = getNonStandardAttributeKeys(log);

		try {
			Set<String> attributesToMove = ImmutableSet.copyOf(ProMUIHelper.queryForObjects(context,
					"Select attributes to move to trace level. The last value is written at trace level.", attributeKeys));
			
			Progress progBar = context.getProgress();
			progBar.setMinimum(0);
			progBar.setMaximum(log.size());
			progBar.setValue(0);

			HashMap<String,XAttribute> map=new HashMap<String, XAttribute>(); 

			for(XTrace trace : log)
			{
				for(XEvent event : trace)
				{
					for(String attributeToMove : attributesToMove)
					{
						XAttribute attr = event.getAttributes().get(attributeToMove);
						if (attr!=null)
							map.put(attributeToMove, attr);
					}
				}
				for(Entry<String, XAttribute> entry : map.entrySet())
				{
					trace.getAttributes().put(entry.getKey(), (XAttribute) entry.getValue().clone());
				}
				map.clear();
				if (progBar!=null)
					progBar.inc();

			}
			
		} catch (UserCancelledException e) {
			System.out.println("User Interrupted");
		}
		
	}	

	@Plugin(name = "Move trace level attributes from events to trace (in place)", parameterLabels = { "Event Log" }, //			
			returnLabels = {}, returnTypes = {}, categories = {
					PluginCategory.Enhancement }, level = PluginLevel.Regular, userAccessible = true, mostSignificantResult = -1, 
							help = "Moves all attributes that appear in every event with the same value to the trace level, i.e., creates an attribute with that value for the trace. "
							+ "Optionally it is possible to move attributes to the trace without checking whether they appear on all events with the same value.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void moveAttributesToTrace(UIPluginContext context, XLog log) throws UserCancelledException {

		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);

		List<String> attributeKeys = getNonStandardAttributeKeys(log);

		Set<String> attributeToCheck = ImmutableSet.copyOf(ProMUIHelper.queryForObjects(context,
				"Select attributes to move when always written with same value", attributeKeys));
		
		attributeKeys.removeAll(attributeKeys);

		Set<String> attributesToMove = ImmutableSet.copyOf(ProMUIHelper.queryForObjects(context,
				"Select attributes to move no matter what value is written", attributeKeys));

		doMoveAttributesToTrace(log, attributeToCheck, attributesToMove, progBar);

	}

	@Plugin(name = "Move trace level attributes from events to trace (not in place)", parameterLabels = { "Event Log" }, //			
			returnLabels = { "Event Log" }, returnTypes = { XLog.class }, categories = {
					PluginCategory.Enhancement }, level = PluginLevel.Regular, userAccessible = true, mostSignificantResult = -1, help = "Moves all attributes that appear in every event with the same value to the trace level, i.e., creates an attribute with that value for the trace. "
							+ "Optionally it is possible to move attributes to the trace without checking whether they appear on all events with the same value.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog moveAttributesToTraceCopy(UIPluginContext context, XLog log) throws UserCancelledException {

		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);

		List<String> attributeKeys = getNonStandardAttributeKeys(log);

		Set<String> attributeToCheck = ImmutableSet.copyOf(ProMUIHelper.queryForObjects(context,
				"Select attributes to move when always written with same value", attributeKeys));

		Set<String> attributesToMove = ImmutableSet.copyOf(ProMUIHelper.queryForObjects(context,
				"Select attributes to move no matter what value is written", attributeKeys));

		XLog changedLog = (XLog) log.clone();
		doMoveAttributesToTrace(changedLog, attributeToCheck, attributesToMove, progBar);
		return changedLog;
	}

	private List<String> getNonStandardAttributeKeys(XLog log) {
		Set<String> eventAttributeKeys = XUtils.getEventAttributeKeys(log);

		List<String> attributeKeys = new ArrayList<>();
		for (String key : eventAttributeKeys) {
			if (!XLifecycleExtension.KEY_TRANSITION.equals(key) && !XLifecycleExtension.KEY_MODEL.equals(key)
					&& !XTimeExtension.KEY_TIMESTAMP.equals(key) && !XConceptExtension.KEY_NAME.equals(key)
					&& !XConceptExtension.KEY_INSTANCE.equals(key) && !XOrganizationalExtension.KEY_ROLE.equals(key)
					&& !XOrganizationalExtension.KEY_RESOURCE.equals(key)
					&& !XOrganizationalExtension.KEY_GROUP.equals(key)) {
				attributeKeys.add(key);
			}
		}
		Collections.sort(attributeKeys);
		return attributeKeys;
	}

	public void doMoveAttributesToTrace(XLog log, Set<String> attributeToCheck, Set<String> attributesToMove,
			Progress progress) {
		ListIterator<XTrace> traceIterator = log.listIterator();

		while (traceIterator.hasNext()) {

			XTrace trace = traceIterator.next();
			ListIterator<XEvent> eventIterator = trace.listIterator();

			Map<String, ValueCounter> attributeOccurence = buildOccurenceMap(eventIterator);

			int eventsInTrace = trace.size();

			for (Entry<String, ValueCounter> occurenceEntry : attributeOccurence.entrySet()) {

				if ((occurenceEntry.getValue().getCounter() == eventsInTrace
						&& attributeToCheck.contains(occurenceEntry.getKey()))
						|| attributesToMove.contains(occurenceEntry.getKey())) {
					// All events share the same attribute with the same value
					moveAttribute(trace, occurenceEntry);
				}
			}

			if (progress != null) {
				progress.inc();
			}
		}
	}

	public void doMoveAttributesToTrace(XLog log, Set<String> attributeToCheck, Progress progress) {
		doMoveAttributesToTrace(log, attributeToCheck, ImmutableSet.<String>of(), progress);
	}

	private static void moveAttribute(XTrace trace, Entry<String, ValueCounter> occurenceEntry) {
		// Add Attribute to Trace
		trace.getAttributes().put(occurenceEntry.getKey(), occurenceEntry.getValue().getValue());

		Iterator<XEvent> eventIterator = trace.iterator();

		while (eventIterator.hasNext()) {
			XEvent event = eventIterator.next();
			// Remove Attribute from all Events
			Iterator<String> attributeNameIter = event.getAttributes().keySet().iterator();
			while (attributeNameIter.hasNext()) {
				String commonAttributeName = occurenceEntry.getKey();
				if (commonAttributeName.equals(attributeNameIter.next())) {
					attributeNameIter.remove();
				}
			}
		}
	}

	private static Map<String, ValueCounter> buildOccurenceMap(ListIterator<XEvent> eventIterator) {
		Map<String, ValueCounter> attributeOccurence = new HashMap<>();

		while (eventIterator.hasNext()) {
			XEvent event = eventIterator.next();

			for (XAttribute attribute : event.getAttributes().values()) {
				String key = attribute.getKey();
				ValueCounter occurenceCount = attributeOccurence.get(key);
				if (occurenceCount == null) {
					attributeOccurence.put(key, new ValueCounter(attribute));
				} else {
					occurenceCount.inc(attribute);
				}
			}
		}
		return attributeOccurence;
	}

}
