package org.processmining.logenhancement.transforming;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

import com.google.common.base.Function;
import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;

public class ObfuscateNumericAttributeIntoBinsPlugin {

	private static final int DEFAULT_NUM_BINS = 10;

	@Plugin(name = "Anonymize: Obfuscate numeric attribute into bins", parameterLabels = { "Event Log" }, //
			returnLabels = { "Obfuscated Log" }, returnTypes = {
					XLog.class }, userAccessible = true, mostSignificantResult = 1, categories = {
							PluginCategory.Enhancement }, help = "Obfuscates numeric attributes (discrete, continous) in a number of bins, which can serve as a way of anonymization. For example: 2 -> 1-20, 31 -> 21-40.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog obfuscateNumericAttributes(UIPluginContext context, XLog log) throws UserCancelledException {

		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, classifier);

		List<XAttribute> attributeCandidates = new ArrayList<>();
		attributeCandidates.addAll(logInfo.getTraceAttributeInfo().getAttributesForType(XAttributeContinuous.class));
		attributeCandidates.addAll(logInfo.getTraceAttributeInfo().getAttributesForType(XAttributeDiscrete.class));
		attributeCandidates.addAll(logInfo.getEventAttributeInfo().getAttributesForType(XAttributeContinuous.class));
		attributeCandidates.addAll(logInfo.getEventAttributeInfo().getAttributesForType(XAttributeDiscrete.class));

		List<String> attributeNames = ProMUIHelper.queryForObjects(context, "Select attributes",
				Iterables.transform(attributeCandidates, new Function<XAttribute, String>() {

					public String apply(XAttribute attr) {
						return attr.getKey();
					}
				}));

		Map<String, Integer> defaultBinWidth = calculateDefaultBinWidth(context, log, attributeNames);

		Map<String, String> binWidthMap = ProMUIHelper.queryMapToString(context, "Specify bin width for each attribute",
				attributeNames, Maps.transformValues(defaultBinWidth, new Function<Integer, String>() {

					public String apply(Integer val) {
						return val.toString();
					}
				}));

		XFactory factory = XFactoryRegistry.instance().currentDefault();

		XLog transformedLog = obfuscateNumericValuesIntoBins(log, attributeNames, binWidthMap, factory);

		return transformedLog;
	}

	public XLog obfuscateNumericValuesIntoBins(XLog log, List<String> attributeNames, Map<String, String> binWidthMap,
			XFactory factory) {
		XLog transformedLog = (XLog) log.clone();
		for (String attribute : attributeNames) {

			//TODO don't do this twice
			SortedSet<Number> availableValues = new TreeSet<Number>();
			for (XTrace t : log) {
				XAttribute attr = t.getAttributes().get(attribute);
				if (attr != null) {
					availableValues.add((Number) XUtils.getAttributeValue(attr));
				}
				for (XEvent e : t) {
					attr = e.getAttributes().get(attribute);
					if (attr != null) {
						availableValues.add((Number) XUtils.getAttributeValue(attr));
					}
				}
			}

			Number minValue = availableValues.first();
			Number maxValue = availableValues.last();

			int widthBin = Integer.parseInt(binWidthMap.get(attribute));

			for (XTrace t : transformedLog) {
				XAttribute attr = t.getAttributes().get(attribute);
				if (attr != null) {
					t.getAttributes().put(attr.getKey(), factory.createAttributeDiscrete(attr.getKey(),
							getBinValue(minValue, maxValue, widthBin, attr), null));
				}
				for (XEvent e : t) {
					attr = e.getAttributes().get(attribute);
					if (attr != null) {
						e.getAttributes().put(attr.getKey(), factory.createAttributeDiscrete(attr.getKey(),
								getBinValue(minValue, maxValue, widthBin, attr), null));
					}
				}
			}

		}
		return transformedLog;
	}

	private Map<String, Integer> calculateDefaultBinWidth(UIPluginContext context, XLog log,
			List<String> attributeNames) {
		Map<String, Integer> defaultBinWidths = new HashMap<String, Integer>();
		for (String attribute : attributeNames) {

			SortedSet<Number> availableValues = new TreeSet<Number>();
			for (XTrace t : log) {
				XAttribute attr = t.getAttributes().get(attribute);
				if (attr != null) {
					availableValues.add((Number) XUtils.getAttributeValue(attr));
				}
				for (XEvent e : t) {
					attr = e.getAttributes().get(attribute);
					if (attr != null) {
						availableValues.add((Number) XUtils.getAttributeValue(attr));
					}
				}
			}

			Number minValue = availableValues.first();
			Number maxValue = availableValues.last();

			int widthBin = (int) Math
					.ceil(Math.max(1d, (maxValue.doubleValue() - minValue.doubleValue()) / DEFAULT_NUM_BINS));
			defaultBinWidths.put(attribute, widthBin);
		}
		return defaultBinWidths;
	}

	private Long getBinValue(Number minValue, Number maxValue, int widthBin, XAttribute attr) {
		if (widthBin > 0) {
			if (attr instanceof XAttributeDiscrete) {
				long value = ((XAttributeDiscrete) attr).getValue();
				return (value / widthBin) * widthBin;
			} else if (attr instanceof XAttributeContinuous) {
				double value = ((XAttributeContinuous) attr).getValue();
				return ((long) Math.ceil(value / widthBin)) * widthBin;
			} else {
				throw new RuntimeException("Undefined for non-numeric value!");
			}
		} else {
			return 0l;
		}
	}

}
