package org.processmining.logenhancement.transforming;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;

public final class RemoveDuplicateAttributeValuesPlugin {

	@Plugin(name = "Remove Duplicate Attribute Values (In-place)", parameterLabels = {}, //
			returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1, //
			categories = { PluginCategory.Enhancement }, keywords = { "remove", "attribute", "duplicate", "repeated" }, //
			help = "Removes duplicate attribute values from events, i.e., attribute that have not change since the last event occurred. "
					+ "This creates a smaller sparse event log, which is useful to reduce the memory consumption.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void removeDuplicateAttributeValues(UIPluginContext context, XLog log) throws UserCancelledException {

		Set<String> relevantAttributes = new HashSet<>(ProMUIHelper.queryForStrings(context,
				"Select which attributes should be considered", XUtils.getEventAttributeKeys(log)));

		doMergeSubsequentEvents(context.getProgress(), log, relevantAttributes);
	}

	@Plugin(name = "Remove Duplicate Attribute Values", parameterLabels = {}, //
			returnLabels = { "Merged Log" }, returnTypes = {
					XLog.class }, userAccessible = true, mostSignificantResult = 1, //
			categories = { PluginCategory.Enhancement }, keywords = { "remove", "attribute", "duplicate", "repeated" }, //
			help = "Removes duplicate attribute values from events, i.e., attribute that have not change since the last event occurred. "
					+ "This creates a smaller sparse event log, which is useful to reduce the memory consumption.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog removeDuplicateAttributeValuesCopy(UIPluginContext context, XLog log) throws UserCancelledException {

		Set<String> relevantAttributes = new HashSet<>(ProMUIHelper.queryForStrings(context,
				"Select which attributes should be considered", XUtils.getEventAttributeKeys(log)));

		return doMergeSubsequentEvents(context.getProgress(), (XLog) log.clone(), relevantAttributes);
	}

	public static XLog doMergeSubsequentEvents(Progress progress, XLog log, Set<String> relevantAttributes) {

		progress.setMaximum(log.size());

		for (XTrace t : log) {
			removeDuplicatesFromTrace(relevantAttributes, t);
			progress.inc();
		}

		return log;
	}

	private static void removeDuplicatesFromTrace(Set<String> relevantAttributes, XTrace t) {
		Map<String, XAttribute> attributeValues = initialAttributes(t, relevantAttributes);
		for (XEvent e : t) {
			for (Iterator<XAttribute> iter = e.getAttributes().values().iterator(); iter.hasNext();) {
				XAttribute newAttribute = iter.next();
				if (relevantAttributes.contains(newAttribute.getKey())) {
					XAttribute lastAttribute = attributeValues.get(newAttribute.getKey());
					if (lastAttribute != null && newAttribute.equals(lastAttribute)) {
						iter.remove();
					} else {
						attributeValues.put(newAttribute.getKey(), newAttribute);
					}
				}
			}
		}
	}

	private static Map<String, XAttribute> initialAttributes(XAttributable attributable,
			Set<String> relevantAttributes) {
		Map<String, XAttribute> attributeValues = new HashMap<>();
		for (XAttribute attr : attributable.getAttributes().values()) {
			if (relevantAttributes.contains(attr.getKey())) {
				attributeValues.put(attr.getKey(), attr);
			}
		}
		return attributeValues;
	}
}