package org.processmining.logenhancement.transforming.classifier;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.DefaultListModel;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableModel;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.util.ui.widgets.ProMList;
import org.processmining.framework.util.ui.widgets.ProMTable;
import org.processmining.log.utils.XUtils;

import com.fluxicon.slickerbox.colors.SlickerColors;
import com.fluxicon.slickerbox.factory.SlickerFactory;

import info.clearthought.layout.TableLayout;
import info.clearthought.layout.TableLayoutConstants;

/**
 * Parameter dialog for the activity mapping of an event log.
 * 
 * @author Eduardo González López de Murillas
 * 
 */
public class ActivityMappingDialog extends JPanel {

	/**
	 * 
	 */
	private static final long serialVersionUID = 7639879370139576539L;

	private DefaultTableModel dm;
	private ActivityMappingParameters parameters;
	private XLog log;
	private JTable mappingTable;
	private JProgressBar progressBar;

	/**
	 * Parameter dialog for selecting the right classifier and mapping.
	 * 
	 * @param log
	 *            The given event log.
	 * @param parameters
	 *            The parameters which will be used for the transformation.
	 */
	public ActivityMappingDialog(XLog l, ActivityMappingParameters p) {

		this.parameters = p;
		this.log = l;

		this.progressBar = new JProgressBar(JProgressBar.HORIZONTAL);
		progressBar.setMaximum(100);
		progressBar.setMinimum(0);
		progressBar.setPreferredSize(new Dimension(250, 20));

		/*
		 * Get a layout containing a single column and three rows, where the top
		 * row height equals 30.
		 */
		double size[][] = { { TableLayoutConstants.FILL },
				{ 30, 20, 120, 5, 20, 30, 20, TableLayoutConstants.FILL, 30 } };
		setLayout(new TableLayout(size));

		/*
		 * Put a meaningful text in the top row.
		 */
		add(SlickerFactory.instance().createLabel("<html><h2>Select transformation parameters</h2>"), "0, 0");
		add(SlickerFactory.instance().createLabel("<html><p>Choose a classifier</p>"), "0, 1");
		/*
		 * Put a list containing all available classifiers in the bottom row.
		 */

		resetTableModel();

		DefaultListModel<XEventClassifier> classifierListModel = new DefaultListModel<>();

		for (XEventClassifier c : XUtils.getStandardAndLogDefinedEventClassifiers(log)) {
			classifierListModel.addElement(c);
		}

		final ProMList<XEventClassifier> classifierList = new ProMList<>("Classifiers", classifierListModel);

		/*
		 * Create a selection listener on this list that changes the parameters
		 * accordingly.
		 */

		classifierList.addListSelectionListener(new ListSelectionListener() {

			public void valueChanged(ListSelectionEvent e) {
				XEventClassifier ec = classifierList.getSelectedValuesList().get(0);

				mappingTable.setEnabled(false);
				progressBar.setValue(0);

				parameters.setClassifier(ec);
				SwingWorker<Void, Void> worker = new SwingWorker<Void, Void>() {
					@Override
					public Void doInBackground() {
						resetTableModel();
						fillTable(log, dm, parameters);
						return null;
					}

					@Override
					public void done() {
						//progressBar.setValue(100);
						progressBar.setValue(100);
						mappingTable.setEnabled(true);
					}
				};

				worker.execute();

			}
		});

		classifierList.setPreferredSize(new Dimension(250, 120));

		add(classifierList, "0, 2");

		add(progressBar, "0, 4");

		JPanel newClassifierNamePanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 5));
		newClassifierNamePanel.setBackground(this.getBackground());
		newClassifierNamePanel.setForeground(this.getForeground());

		final JTextField newClassifierNameField = new JTextField(parameters.getNewClassifierName());

		newClassifierNameField.setBackground(SlickerColors.COLOR_BG_3);
		newClassifierNameField.setForeground(SlickerColors.COLOR_FG);
		newClassifierNameField.setSelectionColor(SlickerColors.COLOR_BG_1);
		newClassifierNameField.setBorder(BorderFactory.createLineBorder(Color.black));

		newClassifierNameField.getDocument().addDocumentListener(new DocumentListener() {

			public void removeUpdate(DocumentEvent e) {
				update();
			}

			public void insertUpdate(DocumentEvent e) {
				update();
			}

			public void changedUpdate(DocumentEvent e) {
				update();
			}

			private void update() {
				parameters.setNewClassifierName(newClassifierNameField.getText());
			}
		});

		JLabel newClassifierNameLabel = SlickerFactory.instance().createLabel("Choose a name for the new classifier: ");

		newClassifierNamePanel.add(newClassifierNameLabel);
		newClassifierNamePanel.add(newClassifierNameField);

		add(newClassifierNamePanel, "0, 5");

		add(SlickerFactory.instance().createLabel("<html><p>Assign the new activity names</p>"), "0, 6");

		Color myBgColor = new Color(150, 150, 150, 255);
		Color myFgColor = new Color(60, 60, 60, 255);
		Color myBdColor = new Color(40, 40, 40, 255);
		progressBar.setBackground(myBgColor);
		progressBar.setForeground(myFgColor);
		progressBar.setBorder(BorderFactory.createLineBorder(myBdColor, 2, false));
		newClassifierNamePanel.setBackground(myBgColor);

		ProMTable promTable = new ProMTable(dm);
		promTable.setPreferredSize(new Dimension(250, 240));
		mappingTable = promTable.getTable();
		add(promTable, "0, 7");

		add(SlickerFactory.instance()
				.createLabel("<html><p><b>NOTE:</b> Empty cells will keep the original activity name</p>"), "0, 8");
	}

	private void resetTableModel() {

		dm = new DefaultTableModel() {
			/**
			 * 
			 */
			private static final long serialVersionUID = -8584494678320655550L;

			public boolean isCellEditable(int row, int column) {
				if (column == 0) {
					return false;
				}
				return super.isCellEditable(row, column);
			}
		};

		dm.setColumnIdentifiers(new String[] { "Original Activity Name", "New Activity Name" });

		dm.addTableModelListener(new TableModelListener() {

			public void tableChanged(TableModelEvent e) {
				Map<String, String> mapping = parameters.getActivityMapping();

				int start = e.getFirstRow();
				int end = e.getLastRow();

				for (int i = start; i <= end; i++) {
					String k = (String) dm.getValueAt(i, 0);
					String v = (String) dm.getValueAt(i, 1);
					if (v != null && v.trim().isEmpty()) {
						v = null;
					}
					mapping.put(k, v);
				}
			}
		});

		if (mappingTable != null) {
			mappingTable.setModel(dm);
		}
	}

	private void fillTable(XLog log, DefaultTableModel dm, ActivityMappingParameters parameters) {
		Map<String, String> mapping = parameters.getActivityMapping();
		Set<String> activities = new HashSet<>();
		XEventClassifier classifier = parameters.getClassifier();

		int traces = log.size();
		int i = 0;
		final int[] prog = new int[] { 0 };

		for (XTrace t : log) {
			for (XEvent e : t) {
				String a = classifier.getClassIdentity(e);
				activities.add(a);
			}
			i++;
			prog[0] = (int) (((float) i / (float) traces) * 95);
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					progressBar.setValue(prog[0]);
				}
			});
		}

		List<String> sortedActivities = new ArrayList<String>(activities);
		Collections.sort(sortedActivities);

		parameters.setActivityMapping(mapping);

		for (String k : sortedActivities) {
			dm.addRow(new String[] { k, mapping.get(k) });
		}

	}

}
