package org.processmining.logenhancement.view;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;

import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.DefaultTableModel;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.util.XAttributeUtils;
import org.processmining.framework.util.ui.widgets.ProMScrollPane;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.ProMTableWithoutPanel;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList.ClickListener;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList.TraceBuilder;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.framework.util.ui.widgets.traceview.masterdetail.DetailView;
import org.processmining.plugins.DataConformance.visualization.grouping.TransparentSplitPane;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Ordering;

public class LogViewDetail extends JPanel implements DetailView<XTrace> {

	private static final String HEADER_TEXT = "TRACE(S) IN SELECTED VARIANTS(S)";

	private static final long serialVersionUID = 5236377082552892703L;

	private static final DateFormat DATE_FORMATTER = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");

	private static final class SelectionWedgeBuilder extends ProMTraceList.DefaultWedgeBuilder {

		private final Event selectedDummyEvent;

		private SelectionWedgeBuilder(Event selectedDummyEvent) {
			this.selectedDummyEvent = selectedDummyEvent;
		}

		public Stroke buildBorderStroke(Trace<? extends Event> trace, Event e) {
			if (e.equals(selectedDummyEvent)) {
				return new BasicStroke(2, BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND, 1f, new float[] { 5f }, 0f);
			}
			return super.buildBorderStroke(trace, e);
		}
	}

	static class EventAttributesPanel extends JPanel {

		private static final long serialVersionUID = 1L;

		private final JTable attributesTable;

		private final DefaultTableModel attributesTableModel;

		private static final Ordering<XAttribute> ATTRIBUTE_ORDER = new Ordering<XAttribute>() {

			public int compare(XAttribute arg0, XAttribute arg1) {
				return arg0.getKey().compareTo(arg1.getKey());
			}
		};

		public EventAttributesPanel() {
			super();
			setBackground(Color.WHITE);
			setForeground(null);
			setOpaque(false);
			setLayout(new BorderLayout());
			attributesTableModel = new DefaultTableModel(new String[] { "Name", "Type", "Value" }, 0) {

				private static final long serialVersionUID = 1L;

				public boolean isCellEditable(int row, int column) {
					return true;
				}
				
			};
			attributesTable = new ProMTableWithoutPanel(getAttributesTableModel());
			attributesTable.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);			
			ProMScrollPane scrollPane = new ProMScrollPane(getAttributesTable());
			scrollPane.setBackground(Color.WHITE);
			scrollPane.setForeground(null);
			scrollPane.setOpaque(false);
			scrollPane.getViewport().setBackground(null);
			scrollPane.getViewport().setForeground(null);
			scrollPane.getViewport().setOpaque(false);
			add(scrollPane, BorderLayout.CENTER);
		}

		public void displayAttributes(XAttributable attributable) {
			getAttributesTableModel().getDataVector().clear();
			getAttributesTable().setModel(getAttributesTableModel());
			attributesTable.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
			ImmutableList<XAttribute> attributes = ATTRIBUTE_ORDER.immutableSortedCopy(attributable.getAttributes()
					.values());
			for (XAttribute a : attributes) {
				if (a instanceof XAttributeLiteral) {
					getAttributesTableModel().addRow(
							new String[] { a.getKey(), XAttributeUtils.getTypeString(a),
									((XAttributeLiteral) a).getValue() });
				} else if (a instanceof XAttributeTimestamp) {
					getAttributesTableModel().addRow(
							new String[] { a.getKey(), XAttributeUtils.getTypeString(a),
									DATE_FORMATTER.format(((XAttributeTimestamp) a).getValue()) });
				} else if (a instanceof XAttributeDiscrete) {
					getAttributesTableModel().addRow(
							new String[] { a.getKey(), XAttributeUtils.getTypeString(a),
									String.valueOf(((XAttributeDiscrete) a).getValue()) });
				} else if (a instanceof XAttributeContinuous) {
					getAttributesTableModel().addRow(
							new String[] { a.getKey(), XAttributeUtils.getTypeString(a),
									String.valueOf(((XAttributeContinuous) a).getValue()) });
				} else if (a instanceof XAttributeBoolean) {
					getAttributesTableModel().addRow(
							new String[] { a.getKey(), XAttributeUtils.getTypeString(a),
									String.valueOf(((XAttributeBoolean) a).getValue()) });
				} else {
					getAttributesTableModel().addRow(
							new String[] { a.getKey(), XAttributeUtils.getTypeString(a), "N/A" });
				}
			}
			getAttributesTable().repaint();
		}

		public JTable getAttributesTable() {
			return attributesTable;
		}

		public DefaultTableModel getAttributesTableModel() {
			return attributesTableModel;
		}

		public void clear() {
			getAttributesTableModel().getDataVector().clear();
			getAttributesTable().repaint();
		}

	}

	private final JLabel headerLabel;
	private final XTraceListView detailView;
	private final EventAttributesPanel eventAttributesPanel;

	public LogViewDetail(LogViewModel logViewModel) {
		super();

		setOpaque(true);
		setBackground(Color.WHITE);
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		// Might be accessed by createDetailView to set the header!
		headerLabel = createHeading(HEADER_TEXT);

		detailView = createDetailView(Collections.<XTrace>emptyList(), logViewModel);
		eventAttributesPanel = new EventAttributesPanel();

		TransparentSplitPane splitPane = new TransparentSplitPane(ProMSplitPane.VERTICAL_SPLIT);
		splitPane.setBackground(Color.WHITE);

		JPanel detailViewPanel = new JPanel();
		detailViewPanel.setOpaque(true);
		detailViewPanel.setBackground(Color.WHITE);
		detailViewPanel.setLayout(new BoxLayout(detailViewPanel, BoxLayout.Y_AXIS));
		detailViewPanel.add(headerLabel);
		detailViewPanel.add(detailView);

		splitPane.setTopComponent(detailViewPanel);
		splitPane.setBottomComponent(eventAttributesPanel);
		splitPane.setResizeWeight(0.8d);
		splitPane.setOneTouchExpandable(true);

		add(splitPane);
	}

	private XTraceListView createDetailView(Collection<XTrace> traces, final LogViewModel logViewModel) {
		final XTraceListView detailListView = new XTraceListView(traces, new TraceBuilder<XTrace>() {

			public Trace<? extends Event> build(XTrace trace) {
				return new XESTraceColored(trace, logViewModel);
			}
		}, null) {

			private static final long serialVersionUID = 1L;

			public void afterUpdate() {
				super.afterUpdate();
				if (getListModel().getSize() > 0) {
					headerLabel.setText(getListModel().getSize() + " " + HEADER_TEXT);
				}
			}

		};
		detailListView.setPreferredSize(new Dimension(400, Short.MAX_VALUE)); //TODO why is that necessary??
		detailListView.setOpaque(true);
		detailListView.setWedgeGap(3);
		detailListView.setMaxWedgeWidth(150);
		detailListView.setCollapsedLabelLength(3);
		detailListView.setFixedWedgeLimit(5000);
		detailListView.setBackground(Color.WHITE);
		detailListView.setForeground(Color.BLACK);
		detailListView.addTraceClickListener(new ClickListener<XTrace>() {

			public void traceMouseDoubleClicked(XTrace trace, int traceIndex, int eventIndex, MouseEvent e) {

			}

			public void traceMouseClicked(XTrace trace, int traceIndex, final int eventIndex, MouseEvent e) {
				if (eventIndex != -1) {
					getAttributesPanel().displayAttributes(trace.get(eventIndex));
					//TODO this should be solved somehow in an easier way
					final Event selectedDummyEvent = new XESEvent(trace.get(eventIndex).getID(), null, null, null,
							null, null);
					ProMTraceList.DefaultWedgeBuilder selectionWedgeBuilder = new SelectionWedgeBuilder(
							selectedDummyEvent);
					detailListView.setWedgeBuilder(selectionWedgeBuilder);
					detailListView.repaint();
				} else {
					getAttributesPanel().displayAttributes(trace);
					detailListView.setWedgeBuilder(null);
					detailListView.repaint();
				}
			}
		});
		return detailListView;
	}

	private JLabel createHeading(String label) {
		JLabel title = new JLabel(label);
		title.setAlignmentX(Component.CENTER_ALIGNMENT);
		title.setFont(title.getFont().deriveFont(Font.BOLD));
		title.setForeground(null);
		return title;
	}

	public JComponent getDetailComponent() {
		return this;
	}

	public ProMTraceList<XTrace> getDetailList() {
		return detailView;
	}

	public Comparator<XTrace> getSortOrder() {
		return new Comparator<XTrace>() {

			public int compare(XTrace o1, XTrace o2) {
				return XConceptExtension.instance().extractName(o1)
						.compareTo(XConceptExtension.instance().extractName(o2));
			}
		};
	}

	public EventAttributesPanel getAttributesPanel() {
		return eventAttributesPanel;
	}

}