package org.processmining.logenhancement.view;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.HeadlessException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.text.DateFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.SwingWorker;
import javax.swing.table.DefaultTableModel;

import org.deckfour.xes.classification.XEventAttributeClassifier;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.util.ui.widgets.ProMComboBox;
import org.processmining.framework.util.ui.widgets.ProMTableWithoutPanel;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList.ClickListener;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList.TraceBuilder;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.framework.util.ui.widgets.traceview.masterdetail.MasterView;
import org.processmining.framework.util.ui.widgets.traceview.masterdetail.TransparentSplitPane;
import org.processmining.log.utils.XUtils;

import com.google.common.base.Function;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableListMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSet.Builder;
import com.google.common.collect.Lists;
import com.google.common.collect.Multiset;
import com.google.common.collect.Multiset.Entry;

/**
 * Master view on the right
 * 
 * @author F. Mannhardt
 * 
 */
public class LogViewMaster extends JPanel
		implements MasterView<Multiset.Entry<XESTraceVariant>, XTrace>, ClickListener<Multiset.Entry<XESTraceVariant>> {

	private static final DateFormat DATE_FORMATTER = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");

	private final class LogInfoWorker extends SwingWorker<LogInfo, LogInfo> {

		protected void process(List<LogInfo> chunks) {
			updateUI(chunks.get(0));
		}

		protected LogInfo doInBackground() throws Exception {
			LogInfo logInfo = new LogInfo();

			Set<XEventClass> eventClasses = new HashSet<>();

			logInfo.numVariants = logViewModel.getTraceVariants().keySet().size();
			
			for (XTrace trace : logViewModel.getTraces()) {
				for (XEvent event : trace) {
					eventClasses.add(logViewModel.getEventClasses().getClassOf(event));
				}
				logInfo.numTraces++;
				logInfo.numEvents += trace.size();
				logInfo.avgTraceLength = logInfo.numEvents / (double) logInfo.numTraces;
				logInfo.numEventClasses = eventClasses.size();
				if (isCancelled()) {
					return logInfo;
				}
				if (logInfo.numTraces % 1000 == 0) {
					publish(logInfo);
				}
			}
			
			publish(logInfo);

			Set<String> attributes = new HashSet<>();
			long firstEvent = Long.MAX_VALUE;
			long lastEvent = Long.MIN_VALUE;
			
			for (XTrace trace : logViewModel.getTraces()) {
				attributes.addAll(trace.getAttributes().keySet());
				for (XEvent event : trace) {
					Date time = XTimeExtension.instance().extractTimestamp(event);
					if (time != null) {
						firstEvent = Math.min(firstEvent, time.getTime());
						lastEvent = Math.max(lastEvent, time.getTime());
					}
					attributes.addAll(event.getAttributes().keySet());
				}
				logInfo.numAttributes = attributes.size();
				if (isCancelled()) {
					return logInfo;
				}
				if (logInfo.numTraces % 1000 == 0) {
					publish(logInfo);
				}
			}
			logInfo.firstEvent = new Date(firstEvent);
			logInfo.lastEvent = new Date(lastEvent);

			return logInfo;
		}

		protected void done() {
			try {
				LogInfo logInfo = get();
				updateUI(logInfo);

			} catch (InterruptedException | CancellationException e) {
				// just return 
			} catch (ExecutionException e) {
				ProMUIHelper.showErrorMessage(mainView, "Error updating log info: " + e.getMessage(), "Log info error");
			}
		}

		private void updateUI(LogInfo logInfo) {
			NumberFormat integerInstance = NumberFormat.getIntegerInstance();
			NumberFormat numberInstance = NumberFormat.getNumberInstance();
			int i = 0;
			infoTableModel.setValueAt(integerInstance.format(logInfo.numTraces), i++, 1);
			infoTableModel.setValueAt(integerInstance.format(logInfo.numEvents), i++, 1);
			infoTableModel.setValueAt(integerInstance.format(logInfo.numEventClasses), i++, 1);
			infoTableModel.setValueAt(integerInstance.format(logInfo.numAttributes), i++, 1);
			infoTableModel.setValueAt(integerInstance.format(logInfo.numVariants), i++, 1);
			infoTableModel.setValueAt(numberInstance.format(logInfo.avgTraceLength), i++, 1);
			if (logInfo.firstEvent != null) {
				infoTableModel.setValueAt(DATE_FORMATTER.format(logInfo.firstEvent), i++, 1);
			}
			if (logInfo.firstEvent != null) {
				infoTableModel.setValueAt(DATE_FORMATTER.format(logInfo.lastEvent), i++, 1);
			}
			logInfoTable.repaint();
		}

	}

	private static final class XEventClassifierUI {

		private final XEventClassifier classifier;

		public XEventClassifierUI(XEventClassifier classifier) {
			this.classifier = classifier;
		}

		public XEventClassifier getClassifier() {
			return classifier;
		}

		public String toString() {
			return "Classify by " + classifier.toString();
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((classifier == null) ? 0 : classifier.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			XEventClassifierUI other = (XEventClassifierUI) obj;
			if (classifier == null) {
				if (other.classifier != null)
					return false;
			} else if (!classifier.equals(other.classifier))
				return false;
			return true;
		}

	}

	private static class LogInfo {

		int numTraces;
		long numEvents;
		int numVariants;

		double avgTraceLength;

		int numEventClasses;
		int numAttributes;

		Date firstEvent;
		Date lastEvent;

	}

	private static final long serialVersionUID = 2L;

	private final LogViewModel logViewModel;
	private TraceGroupingFunction currentGrouping;

	private final ProMTraceList<Entry<XESTraceVariant>> mainView;

	private final DefaultTableModel infoTableModel;

	private JTable logInfoTable;

	private final JComboBox<GroupOrdering> sortComboBox;
	private final JComboBox<GroupingMode> groupComboBox;

	private final JComboBox<ColoringMode> coloringComboBox;
	private final DefaultComboBoxModel<ColoringMode> coloringModel;

	private final JComboBox<XEventClassifierUI> eventClassComboBox;

	private SwingWorker<LogInfo, LogInfo> logInfoWorker;

	@SuppressWarnings("unchecked")
	public LogViewMaster(final LogViewModel logViewModel) {
		super();
		this.logViewModel = logViewModel;
		this.currentGrouping = new TraceGroupingFunctionByExactSequence(logViewModel);

		setOpaque(true);
		setBackground(Color.WHITE);
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		mainView = createMainView(logViewModel.getTraceVariants());

		TransparentSplitPane leftBottomPanel = new TransparentSplitPane(JSplitPane.VERTICAL_SPLIT);
		leftBottomPanel.setBackground(null);

		sortComboBox = new ProMComboBox<GroupOrdering>(GroupOrdering.values()).resetSizes();
		sortComboBox.setSelectedItem(GroupOrdering.COUNT_DESC);
		sortComboBox.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				mainView.sort(getCurrentOrdering());
			}
		});
		sortComboBox.setPreferredSize(new Dimension(50, 30));
		mainView.getToolbar().add(sortComboBox);

		groupComboBox = new ProMComboBox<GroupingMode>(GroupingMode.values()).resetSizes();
		groupComboBox.setSelectedItem(GroupingMode.TOTAL_ORDER);
		groupComboBox.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				GroupingMode mode = getGroupingMode();
				reloadTraces(logViewModel.getTraces(), mode.createGrouping(logViewModel));
			}

		});
		groupComboBox.setPreferredSize(new Dimension(50, 30));
		mainView.getToolbar().add(groupComboBox);

		coloringModel = new DefaultComboBoxModel<ColoringMode>(getDefaultColoringModes());
		coloringComboBox = new ProMComboBox<ColoringMode>(coloringModel).resetSizes();
		coloringComboBox.setSelectedItem(ColoringMode.EVENTCLASS);
		coloringComboBox.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				beforeUpdate();
				new SwingWorker<Void, Void>() {

					protected Void doInBackground() throws Exception {
						logViewModel.setColoring(getColoringMode());
						return null;
					}

					protected void done() {
						afterUpdate();
						super.done();
					}

				}.execute();
			}

		});
		coloringComboBox.setPreferredSize(new Dimension(50, 30));
		mainView.getToolbar().add(coloringComboBox);

		eventClassComboBox = createEventClassComboBox();
		mainView.getToolbar().add(eventClassComboBox);

		infoTableModel = new DefaultTableModel(new String[] { "Name", "Value" }, 0);
		infoTableModel.addRow(new String[] { "Traces", "?" });
		infoTableModel.addRow(new String[] { "Events", "?" });
		infoTableModel.addRow(new String[] { "Event Classes", "?" });
		infoTableModel.addRow(new String[] { "Attributes", "?" });
		infoTableModel.addRow(new String[] { "Variants", "?" });
		infoTableModel.addRow(new String[] { "Events per Trace", "?" });
		infoTableModel.addRow(new String[] { "First Event", "?" });
		infoTableModel.addRow(new String[] { "Last Event", "?" });

		logInfoTable = new ProMTableWithoutPanel(infoTableModel);
		logInfoTable.setBorder(BorderFactory.createMatteBorder(0, 1, 0, 0, Color.BLACK));
		logInfoTable.setBackground(Color.WHITE);

		leftBottomPanel.setTopComponent(mainView);
		leftBottomPanel.setResizeWeight(1);
		leftBottomPanel.setOneTouchExpandable(true);
		leftBottomPanel.setBottomComponent(logInfoTable);
		add(leftBottomPanel);
	}

	private void updateInfoModel() {
		if (logInfoWorker != null && !logInfoWorker.isDone()) {
			logInfoWorker.cancel(false);
		}
		logInfoWorker = new LogInfoWorker();
		logInfoWorker.execute();
	}

	private JComboBox<XEventClassifierUI> createEventClassComboBox() {
		Set<XEventClassifierUI> classifiers;
		if (logViewModel.getAllTraces() instanceof XLog) {
			Builder<XEventClassifierUI> builder = ImmutableSet.builder();
			for (XEventClassifier classifier : getClassifiers((XLog) logViewModel.getAllTraces())) {
				builder.add(new XEventClassifierUI(classifier));
			}
			builder.add(new XEventClassifierUI(logViewModel.getEventClasses().getClassifier()));
			classifiers = builder.build();
		} else {
			classifiers = ImmutableSet.of(new XEventClassifierUI(logViewModel.getEventClasses().getClassifier()));
		}
		@SuppressWarnings("unchecked")
		final JComboBox<XEventClassifierUI> eventClassesBox = new ProMComboBox<XEventClassifierUI>(
				classifiers.toArray()).resetSizes();
		eventClassesBox.setSelectedItem(new XEventClassifierUI(logViewModel.getEventClasses().getClassifier()));
		eventClassesBox.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {

				final XEventClassifier newClassifier = ((XEventClassifierUI) eventClassesBox.getSelectedItem())
						.getClassifier();

				beforeUpdate();

				new SwingWorker<Collection<XTrace>, Void>() {

					protected Collection<XTrace> doInBackground() throws Exception {
						logViewModel.setEventClasses(
								LogViewVisualizer.createEventClasses(logViewModel.getAllTraces(), newClassifier));
						return logViewModel.getAllTraces();
					}

					protected void done() {
						afterUpdate();
						reloadTraces(logViewModel.getTraces());
						super.done();
					}

				}.execute();

			}
		});
		eventClassesBox.setPreferredSize(new Dimension(50, 30));
		return eventClassesBox;
	}

	private List<XEventClassifier> getClassifiers(XLog log) {
		List<XEventClassifier> classList = new ArrayList<>(log.getClassifiers());
		Set<String> globalKeys = ImmutableSet
				.copyOf(Lists.transform(log.getGlobalEventAttributes(), new Function<XAttribute, String>() {

					public String apply(XAttribute attribute) {
						return attribute.getKey();
					}
				}));
		addClassifierIfPossible(XLogInfoImpl.STANDARD_CLASSIFIER, classList, globalKeys);
		addClassifierIfPossible(XLogInfoImpl.NAME_CLASSIFIER, classList, globalKeys);
		addClassifierIfPossible(XLogInfoImpl.RESOURCE_CLASSIFIER, classList, globalKeys);
		for (String key : globalKeys) {
			classList.add(new XEventAttributeClassifier(key, key));
		}
		return classList;
	}

	private void addClassifierIfPossible(XEventClassifier classifier, List<XEventClassifier> classList,
			Set<String> globalKeys) {
		if (!classList.contains(classifier) && hasClassifierAttributes(classifier, globalKeys)) {
			classList.add(classifier);
		}
	}

	private boolean hasClassifierAttributes(XEventClassifier classifier, Set<String> globalKeys) {
		for (String key : classifier.getDefiningAttributeKeys()) {
			if (!globalKeys.contains(key)) {
				return false;
			}
		}
		return true;
	}

	public GroupOrdering getCurrentOrdering() {
		return (GroupOrdering) sortComboBox.getSelectedItem();
	}

	private GroupingMode getGroupingMode() {
		GroupingMode mode = (GroupingMode) groupComboBox.getSelectedItem();
		return mode;
	}

	private ColoringMode[] getDefaultColoringModes() {
		return new ColoringMode[] { ColoringMode.EVENTCLASS, ColoringMode.NONE };
	}

	private ColoringMode getColoringMode() {
		ColoringMode coloringmode = (ColoringMode) coloringComboBox.getSelectedItem();
		if (coloringmode == null) {
			return ColoringMode.EVENTCLASS;
		}
		return coloringmode;
	}

	private final ProMTraceList<Entry<XESTraceVariant>> createMainView(
			final ImmutableListMultimap<XESTraceVariant, XTrace> groupedTraces) {
		final ProMTraceList<Multiset.Entry<XESTraceVariant>> mainView = new ProMTraceList<>(
				groupedTraces.keys().entrySet(), new TraceBuilder<Multiset.Entry<XESTraceVariant>>() {

					public Trace<? extends Event> build(Multiset.Entry<XESTraceVariant> element) {
						XESTraceVariant trace = element.getElement();
						trace.setCount(element.getCount());
						return trace;
					}
				}, GroupOrdering.COUNT_DESC, new Font(Font.SANS_SERIF, Font.PLAIN, 10), false);
		mainView.setOpaque(true);
		mainView.setBackground(Color.WHITE);
		mainView.setForeground(Color.BLACK);
		mainView.addTraceClickListener(this);
		mainView.setWedgeGap(3);
		mainView.setCollapsedLabelLength(3);
		mainView.setPreferredSize(new Dimension(800, Short.MAX_VALUE));
		mainView.showToolbar();
		return mainView;
	}

	public JComponent getMasterComponent() {
		return this;
	}

	public ProMTraceList<Entry<XESTraceVariant>> getMasterList() {
		return mainView;
	}

	public Collection<XTrace> getDetailElements(Entry<XESTraceVariant> element) {
		return logViewModel.getTraceVariants().get(element.getElement());
	}

	public void traceMouseDoubleClicked(final Entry<XESTraceVariant> trace, final int traceIndex, int eventIndex,
			MouseEvent e) {
		SwingWorker<Collection<XTrace>, Object> bgWorker = new SwingWorker<Collection<XTrace>, Object>() {

			@Override
			public Collection<XTrace> doInBackground() {
				return logViewModel.getTraceVariants()
						.get(mainView.getList().getModel().getElementAt(traceIndex).getElement());
			}

			@Override
			protected void done() {
				try {
					Collection<XTrace> selectedTraces = get();
					JFrame newFrame = new JFrame("Detail View (" + selectedTraces.size() + " traces");
					LogViewDetail detailView = new LogViewDetail(logViewModel);
					detailView.getDetailList().addAll(selectedTraces);
					newFrame.getContentPane().add(detailView.getDetailComponent());
					newFrame.pack();
					newFrame.setVisible(true);
				} catch (HeadlessException | InterruptedException | ExecutionException e) {
					ProMUIHelper.showErrorMessage(mainView, "Error loading detail elements " + e.getMessage(),
							"Loading error");
				}
			}

		};
		bgWorker.execute();
	}

	public void traceMouseClicked(Entry<XESTraceVariant> trace, int traceIndex, int eventIndex, MouseEvent e) {
		// no op
	}

	public void reloadTraces(final Iterable<XTrace> traces, TraceGroupingFunction grouping) {
		this.currentGrouping = grouping;
		beforeUpdate();
		mainView.getList().getSelectionModel().clearSelection();
		new SwingWorker<Void, Void>() {

			protected Void doInBackground() throws Exception {
				final ImmutableList<XTrace> tracesCopy = ImmutableList.copyOf(traces);
				logViewModel.setTraces(tracesCopy);
				logViewModel.setColoring(getColoringMode());
				logViewModel.setTraceVariants(XUtils.getVariants(tracesCopy, currentGrouping));
				return null;
			}

			protected void done() {
				try {
					get();

					//Start asynchronous update of the log info
					updateInfoModel();

					coloringModel.removeAllElements();
					coloringModel.addElement(ColoringMode.EVENTCLASS);
					for (final String attribute : logViewModel.getGlobalAttributes()) {
						coloringModel.addElement(new ColoringModeByAttribute(attribute));
					}
					coloringModel.addElement(ColoringMode.NONE);

					mainView.clear();
					mainView.addAll(logViewModel.getTraceVariants().keys().entrySet());
					mainView.sort((GroupOrdering) sortComboBox.getSelectedItem());
				} catch (InterruptedException | ExecutionException e) {
					ProMUIHelper.showErrorMessage(mainView, e.toString(), "Error updating view");
				} finally {
					afterUpdate();
				}
			}

		}.execute();
	}

	public void reloadTraces(Iterable<XTrace> traces) {
		reloadTraces(traces, currentGrouping);
	}

	public TraceGroupingFunction getCurrentGroupingFunction() {
		return currentGrouping;
	}

	private void beforeUpdate() {
		mainView.beforeUpdate();
		groupComboBox.setEnabled(false);
		eventClassComboBox.setEnabled(false);
		coloringComboBox.setEnabled(false);
		sortComboBox.setEnabled(false);
	}

	private void afterUpdate() {
		validate();
		groupComboBox.setEnabled(true);
		eventClassComboBox.setEnabled(true);
		coloringComboBox.setEnabled(true);
		sortComboBox.setEnabled(true);
		mainView.afterUpdate();
	}

}