package org.processmining.logenhancement.view.comparison;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;

import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.util.XAttributeUtils;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.ProMScrollPane;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.ProMTableWithoutPanel;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList.ClickListener;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList.TraceBuilder;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.DefaultTrace;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.log.utils.XUtils;
import org.processmining.logenhancement.view.comparison.LogComparisonPlugin.LogComparisonResult;
import org.processmining.logenhancement.view.comparison.LogComparisonPlugin.TraceEntry;
import org.processmining.plugins.utils.sequencedistance.SequenceDistance.DistanceResult;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.base.Function;
import com.google.common.collect.ComparisonChain;
import com.google.common.collect.Lists;
import com.google.common.collect.Ordering;
import com.google.common.collect.Sets;

public class LogComparisonVisualizer {

	private enum ComparisonOrdering implements Comparator<TraceEntry> {
		DISTANCE_ASC("Distance (Ascending)") {
			@Override
			public int compare(final TraceEntry a, final TraceEntry b) {
				return ComparisonChain.start().compare(a.getDistance(), b.getDistance()).result();
			}
		},
		DISTANCE_DESC("Distance (Descending)") {
			@Override
			public int compare(final TraceEntry a, final TraceEntry b) {
				return Collections.reverseOrder(DISTANCE_ASC).compare(a, b);
			}
		},
		NAME_ASC("Name (Ascending)") {
			@Override
			public int compare(final TraceEntry a, final TraceEntry b) {
				return ComparisonChain.start().compare(a.getName(), b.getName()).result();
			}
		},
		NAME_DESC("Name (Descending)") {
			@Override
			public int compare(final TraceEntry a, final TraceEntry b) {
				return Collections.reverseOrder(NAME_ASC).compare(a, b);
			}
		},
		LENGTH_ASC("Length (Ascending)") {
			@Override
			public int compare(final TraceEntry a, final TraceEntry b) {
				return ComparisonChain.start().compare(a.getMaxLength(), b.getMaxLength()).result();
			}
		},
		LENGTH_DESC("Length (Descending)") {
			@Override
			public int compare(final TraceEntry a, final TraceEntry b) {
				return Collections.reverseOrder(LENGTH_ASC).compare(a, b);
			}
		};

		private String description;

		private ComparisonOrdering(String description) {
			this.description = description;
		}

		@Override
		public String toString() {
			return description;
		}
	}

	private static final Color MISMATCH_COLOR = Color.RED;
	private static final Color MATCH_COLOR = Color.GREEN;

	public class ComparisionTrace extends ProMTraceView.AbstractTrace<Event> {

		private final TraceEntry trace;

		public ComparisionTrace(TraceEntry trace) {
			this.trace = trace;
		}

		public String getName() {
			return trace.getName();
		}

		public String getInfo() {
			return "Distance: ".concat(String.valueOf(trace.getDistance()));
		}

		protected List<Event> delegate() {
			if (trace.getDistanceResult() != null) {
				return Lists.transform(trace.getDistanceResult().getAlignment(),
						new Function<DistanceResult.Entry<XEvent>, Event>() {

							public Event apply(DistanceResult.Entry<XEvent> entry) {
								if (entry.getA() == null) {
									return new ProMTraceView.DefaultSplittedEvent(MISMATCH_COLOR, null, MATCH_COLOR,
											XUtils.getConceptName(entry.getB()));
								} else if (entry.getB() == null) {
									return new ProMTraceView.DefaultSplittedEvent(MATCH_COLOR, XUtils
											.getConceptName(entry.getA()), MISMATCH_COLOR, null);
								} else {
									return new ProMTraceView.DefaultEvent(MATCH_COLOR, XUtils.getConceptName(entry
											.getA()));
								}
							}

						});
			} else {
				DefaultTrace<Event> traceForDisplay = new ProMTraceView.DefaultTrace<>(trace.getName(), "Distance: \u221E", trace.getMaxLength());
				if (trace.getTraceA() == null) {
					for (XEvent e : trace.getTraceB()) {
						traceForDisplay.add(new ProMTraceView.DefaultSplittedEvent(MISMATCH_COLOR, null, MATCH_COLOR, XUtils
								.getConceptName(e)));
					}
				} else {
					for (XEvent e : trace.getTraceA()) {
						traceForDisplay.add(new ProMTraceView.DefaultSplittedEvent(MATCH_COLOR, XUtils.getConceptName(e),
								MISMATCH_COLOR, null));
					}
				}
				return traceForDisplay;
			}
		}

	}

	private static class CompareEventAttributesPanel extends JPanel {

		private static final long serialVersionUID = 1L;

		private final JTable attributesTable;

		private final DefaultTableModel attributesTableModel;

		private static final Ordering<XAttribute> ATTRIBUTE_ORDER = new Ordering<XAttribute>() {

			public int compare(XAttribute arg0, XAttribute arg1) {
				return arg0.getKey().compareTo(arg1.getKey());
			}
		};

		@SuppressWarnings("serial")
		public CompareEventAttributesPanel() {
			super();
			setBackground(null);
			setForeground(null);
			setOpaque(false);
			setLayout(new BorderLayout());
			attributesTableModel = new DefaultTableModel(new String[] { "Name", "Type", "Value A", "Value B" }, 0);
			attributesTable = new ProMTableWithoutPanel(attributesTableModel);
			attributesTable.getColumnModel().getColumn(2).setCellRenderer(new DefaultTableCellRenderer() {

				public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
						boolean hasFocus, int row, int column) {
					Component cell = super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row,
							column);
					cell.setForeground(Color.BLACK);
					TableModel model = table.getModel();
					if (column == 2) {
						if (!model.getValueAt(row, column + 1).equals(value)) {
							colorCell(cell, Color.RED);
						}
					} else if (column == 3) {
						if (!model.getValueAt(row, column - 1).equals(value)) {
							colorCell(cell, Color.RED);
						}
					}
					return cell;
				}

				private void colorCell(Component cell, Color color) {
					cell.setForeground(color);
				}

			});
			ProMScrollPane scrollPane = new ProMScrollPane(attributesTable);
			scrollPane.setBackground(null);
			scrollPane.setForeground(null);
			scrollPane.setOpaque(false);
			scrollPane.getViewport().setBackground(null);
			scrollPane.getViewport().setForeground(null);
			scrollPane.getViewport().setOpaque(false);
			add(scrollPane, BorderLayout.CENTER);
		}

		public void displayAttributes(XAttributable attributableA, XAttributable attributableB) {

			attributesTableModel.getDataVector().removeAllElements();
			attributesTable.setModel(attributesTableModel);

			Set<XAttribute> attributesA = Sets.newTreeSet(ATTRIBUTE_ORDER);
			Set<XAttribute> attributesB = Sets.newTreeSet(ATTRIBUTE_ORDER);
			if (attributableA != null) {
				attributesA.addAll(attributableA.getAttributes().values());
			} else {
				attributesA = Collections.emptySet();
			}
			if (attributableB != null) {
				attributesB.addAll(attributableB.getAttributes().values());
			} else {
				attributesB = Collections.emptySet();
			}

			for (XAttribute a : attributesA) {
				XAttribute b = attributableB != null ? attributableB.getAttributes().get(a.getKey()) : null;
				if (b != null) {
					attributesB.remove(b);
				}
				displayAttribute(a, b);
			}

			for (XAttribute b : attributesB) {
				XAttribute a = attributableA != null ? attributableA.getAttributes().get(b.getKey()) : null;
				if (a != null) {
					attributesA.remove(a);
				}
				displayAttribute(a, b);
			}

			attributesTable.repaint();
		}

		private void displayAttribute(XAttribute a, XAttribute b) {
			if (a == null || b == null || a.getClass() == b.getClass()) {
				if (nullSafe(a, b) instanceof XAttributeLiteral) {
					attributesTableModel.addRow(new String[] { nullSafeKey(a, b),
							XAttributeUtils.getTypeString(a != null ? a : b), nullSafeLiteral(a), nullSafeLiteral(b) });
				} else if (nullSafe(a, b) instanceof XAttributeTimestamp) {
					attributesTableModel.addRow(new String[] { nullSafeKey(a, b),
							XAttributeUtils.getTypeString(a != null ? a : b), nullSafeTimestamp(a),
							nullSafeTimestamp(b) });
				} else if (nullSafe(a, b) instanceof XAttributeDiscrete) {
					attributesTableModel
							.addRow(new String[] { nullSafeKey(a, b), XAttributeUtils.getTypeString(a != null ? a : b),
									nullSafeDiscrete(a), nullSafeDiscrete(b) });
				} else if (nullSafe(a, b) instanceof XAttributeContinuous) {
					attributesTableModel.addRow(new String[] { nullSafeKey(a, b),
							XAttributeUtils.getTypeString(a != null ? a : b), nullSafeContinuous(a),
							nullSafeContinuous(b) });
				} else if (nullSafe(a, b) instanceof XAttributeBoolean) {
					attributesTableModel.addRow(new String[] { nullSafeKey(a, b),
							XAttributeUtils.getTypeString(a != null ? a : b), nullSafeBoolean(a), nullSafeBoolean(b) });
				} else {
					attributesTableModel.addRow(new String[] { nullSafeKey(a, b),
							XAttributeUtils.getTypeString(a != null ? a : b), "N/A" });
				}
			} else {
				attributesTableModel.addRow(new String[] { nullSafeKey(a, b), "Type mismatch",
						XAttributeUtils.getTypeString(a != null ? a : b), XAttributeUtils.getTypeString(b) });
			}
		}

		private XAttribute nullSafe(XAttribute a, XAttribute b) {
			return a != null ? a : b;
		}

		private String nullSafeKey(XAttribute a, XAttribute b) {
			return a != null ? a.getKey() : (b != null ? b.getKey() : "N/A");
		}

		private String nullSafeLiteral(XAttribute a) {
			return a != null ? ((XAttributeLiteral) a).getValue() : "N/A";
		}

		private String nullSafeDiscrete(XAttribute a) {
			return a != null ? String.valueOf(((XAttributeDiscrete) a).getValue()) : "N/A";
		}

		private String nullSafeContinuous(XAttribute a) {
			return a != null ? String.valueOf(((XAttributeContinuous) a).getValue()) : "N/A";
		}

		private String nullSafeBoolean(XAttribute a) {
			return a != null ? String.valueOf(((XAttributeBoolean) a).getValue()) : "N/A";
		}

		private String nullSafeTimestamp(XAttribute a) {
			return a != null ? String.valueOf(((XAttributeTimestamp) a).getValue()) : "N/A";
		}

	}

	private ProMTraceList<TraceEntry> mainView;
	private CompareEventAttributesPanel attributesPanel;

	@Plugin(name = "Log Comparison Visualizer", returnLabels = { "Log Comparison" }, returnTypes = { JComponent.class }, userAccessible = true, parameterLabels = { "Log Comparison Result" })
	@Visualizer
	@PluginVariant(requiredParameterLabels = { 0 })
	public JComponent visualise(PluginContext context, final LogComparisonResult comparisonResult) throws Exception {

		context.getProgress().setCaption("Visualising ...");

		ProMSplitPane splitPane = new ProMSplitPane(ProMSplitPane.VERTICAL_SPLIT);
		splitPane.setResizeWeight(1.0d);

		JPanel topPanel = new JPanel();
		topPanel.setOpaque(true);
		topPanel.setBackground(Color.WHITE);
		topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.Y_AXIS));
		mainView = createMainView(comparisonResult.getTraceEntries());
		topPanel.add(mainView);

		JPanel bottomPanel = new JPanel();
		bottomPanel.setOpaque(true);
		bottomPanel.setBackground(Color.WHITE);
		bottomPanel.setLayout(new BoxLayout(bottomPanel, BoxLayout.X_AXIS));
		bottomPanel.add(createSortComboBox());
		bottomPanel.add(Box.createHorizontalStrut(20));
		attributesPanel = new CompareEventAttributesPanel();
		bottomPanel.add(attributesPanel);

		splitPane.setTopComponent(topPanel);
		splitPane.setBottomComponent(bottomPanel);
		return splitPane;
	}

	@SuppressWarnings("rawtypes")
	private JComboBox createSortComboBox() {
		final JComboBox groupSortBox = SlickerFactory.instance().createComboBox(ComparisonOrdering.values());
		groupSortBox.setSelectedItem(ComparisonOrdering.NAME_ASC);
		groupSortBox.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				ComparisonOrdering ordering = (ComparisonOrdering) groupSortBox.getSelectedItem();
				mainView.sort(ordering);
			}
		});
		groupSortBox.setPreferredSize(new Dimension(150, 30));
		return groupSortBox;
	}

	private ProMTraceList<TraceEntry> createMainView(Collection<TraceEntry> traceEntries) {
		final ProMTraceList<TraceEntry> mainView = new ProMTraceList<>(traceEntries, new TraceBuilder<TraceEntry>() {

			public Trace<? extends Event> build(TraceEntry element) {
				return new ComparisionTrace(element);
			}
		}, ComparisonOrdering.NAME_ASC);
		mainView.setOpaque(true);
		mainView.setBackground(Color.WHITE);
		mainView.setForeground(Color.BLACK);
		mainView.setPreferredSize(new Dimension(800, Short.MAX_VALUE));

		mainView.addTraceClickListener(new ClickListener<TraceEntry>() {

			public void traceMouseDoubleClicked(TraceEntry trace, int traceIndex, int eventIndex, MouseEvent e) {
			}

			public void traceMouseClicked(TraceEntry trace, int traceIndex, final int eventIndex, MouseEvent e) {
				if (eventIndex != -1) {
					if (trace.getDistanceResult() != null) {
						DistanceResult.Entry<XEvent> resultEntry = trace.getDistanceResult().getAlignment()
								.get(eventIndex);
						attributesPanel.displayAttributes(resultEntry.getA(), resultEntry.getB());
					}
				}
			}
		});

		return mainView;
	}

}
